"""Classes for training ivector models"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy import gmm, ivector
from _kalpy.hmm import RandomAccessPosteriorReader
from kalpy.feat.data import FeatureArchive
from kalpy.ivector.data import GselectArchive
from kalpy.utils import read_kaldi_object

logger = logging.getLogger("kalpy.ivector")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class GlobalGmmStatsAccumulator:
    def __init__(self, model_path: typing.Union[pathlib.Path, str]):
        self.model_path = str(model_path)
        self.model: gmm.DiagGmm = read_kaldi_object(gmm.DiagGmm, self.model_path)
        self.gmm_accs = gmm.AccumDiagGmm()
        self.gmm_accs.Resize(self.model, gmm.StringToGmmFlags("mvw"))
        self.num_done = 0
        self.num_skipped = 0

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        gselect_archive: GselectArchive,
        callback: typing.Callable = None,
    ):
        tot_like = 0.0
        tot_t = 0.0
        for utt_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utt_id} due to zero-length features")
                self.num_skipped += 1
                continue
            try:
                gselect = gselect_archive[utt_id]
            except KeyError:
                logger.warning(f"Skipping {utt_id} due to missing gselect")
                self.num_skipped += 1
                continue

            if callback:
                callback(utt_id)
            tot_like_this_file = self.gmm_accs.accumulate_from_gselect(self.model, gselect, feats)
            tot_t += feats.NumRows()
            tot_like += tot_like_this_file
            self.num_done += 1
        logger.info(f"Done {self.num_done} files, skipped {self.num_skipped}.")
        logger.info(
            f"Overall avg like per frame (Gaussian only) = {tot_like/tot_t} over {tot_t} frames."
        )


class IvectorExtractorStatsAccumulator:
    def __init__(self, ivector_extractor_path: typing.Union[pathlib.Path, str]):
        self.ivector_extractor_path = str(ivector_extractor_path)

        self.model: ivector.IvectorExtractor = read_kaldi_object(
            ivector.IvectorExtractor, self.ivector_extractor_path
        )
        self.options = ivector.IvectorExtractorStatsOptions()
        self.ivector_stats = ivector.IvectorExtractorStats(self.model, self.options)
        self.num_done = 0
        self.num_skipped = 0

    def accumulate_stats(
        self,
        feature_archive: FeatureArchive,
        post_reader: RandomAccessPosteriorReader,
        callback: typing.Callable = None,
    ):
        for utt_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utt_id} due to zero-length features")
                self.num_skipped += 1
                continue
            if not post_reader.HasKey(utt_id):
                logger.warning(f"Skipping {utt_id} due to missing posterior")
                self.num_skipped += 1
                continue
            post = post_reader.Value(utt_id)
            if len(post) != feats.NumRows():
                logger.warning(f"Skipping {utt_id} due to mismatch in lengths")
                self.num_skipped += 1
                continue
            self.ivector_stats.AccStatsForUtterance(self.model, feats, post)
            self.num_done += 1
            if callback:
                callback(utt_id)
        logger.info(f"Done {self.num_done} files, skipped {self.num_skipped}.")



import os
import pathlib
import subprocess
from io import BytesIO

import librosa
import pytest
import soundfile


@pytest.fixture(scope="session")
def test_dir():
    base = pathlib.Path(__file__).parent
    return base.joinpath("data")


@pytest.fixture(scope="session")
def wav_dir(test_dir):
    return test_dir.joinpath("wav")


@pytest.fixture(scope="session")
def am_dir(test_dir):
    return test_dir.joinpath("acoustic_models")


@pytest.fixture(scope="session")
def mono_am_dir(am_dir):
    return am_dir.joinpath("mono")


@pytest.fixture(scope="session")
def sat_am_dir(am_dir):
    return am_dir.joinpath("sat")


@pytest.fixture(scope="session")
def english_mfa_am_dir(am_dir):
    return am_dir.joinpath("english_mfa")


@pytest.fixture(scope="session")
def temp_dir(test_dir):
    p = test_dir.joinpath("temp")
    p.mkdir(exist_ok=True)
    return p


@pytest.fixture(scope="session")
def reference_dir(test_dir):
    p = test_dir.joinpath("kaldi")
    return p


@pytest.fixture(scope="session")
def reference_mfcc_path(wav_path, reference_dir):
    ark_path = reference_dir.joinpath("mfccs.ark")
    scp_path = reference_dir.joinpath("mfccs.scp")

    mfcc_proc = subprocess.Popen(
        [
            "compute-mfcc-feats",
            "--use-energy=false",
            "--dither=0",
            "--energy-floor=0",
            "--num-ceps=13",
            "--num-mel-bins=23",
            "--cepstral-lifter=22",
            "--preemphasis-coefficient=0.97",
            "--frame-shift=10",
            "--frame-length=25",
            "--low-freq=20",
            "--high-freq=7800",
            "--sample-frequency=16000",
            "--allow-downsample=true",
            "--allow-upsample=true",
            "--snip-edges=false",
            "ark,s,cs:-",
            f"ark,scp:{ark_path},{scp_path}",
        ],
        stdin=subprocess.PIPE,
        env=os.environ,
    )

    wave, _ = librosa.load(
        wav_path,
        sr=16000,
        offset=0.0,
        duration=26.72325,
        mono=False,
    )
    bio = BytesIO()
    soundfile.write(bio, wave, samplerate=16000, format="WAV")
    mfcc_proc.stdin.write(f"1-1\t".encode("utf8"))
    mfcc_proc.stdin.write(bio.getvalue())
    mfcc_proc.stdin.flush()
    mfcc_proc.stdin.close()
    mfcc_proc.communicate()
    return scp_path


@pytest.fixture(scope="session")
def reference_cmvn_path(wav_path, reference_dir, reference_mfcc_path):
    ark_path = reference_dir.joinpath("cmvn.ark")
    scp_path = reference_dir.joinpath("cmvn.scp")
    spk2utt = reference_dir.joinpath("spk2utt.scp")
    with open(spk2utt, "w", encoding="utf8") as f:
        f.write("1 1-1\n")
    subprocess.call(
        [
            "compute-cmvn-stats",
            f"--spk2utt=ark:{spk2utt}",
            f"scp:{reference_mfcc_path}",
            f"ark,scp:{ark_path},{scp_path}",
        ],
        env=os.environ,
    )
    return scp_path


@pytest.fixture(scope="session")
def reference_final_features_path(
    wav_path, reference_dir, reference_mfcc_path, reference_cmvn_path
):
    ark_path = reference_dir.joinpath("final_features.ark")
    scp_path = reference_dir.joinpath("final_features.scp")
    utt2spk = reference_dir.joinpath("utt2spk.scp")
    with open(utt2spk, "w", encoding="utf8") as f:
        f.write("1-1 1\n")
    subprocess.call(
        [
            "apply-cmvn",
            f"--utt2spk=ark:{utt2spk}",
            f"scp:{reference_cmvn_path}",
            f"scp:{reference_mfcc_path}",
            f"ark,scp:{ark_path},{scp_path}",
        ],
        env=os.environ,
    )
    return scp_path


@pytest.fixture(scope="session")
def reference_si_feature_string(reference_final_features_path, sat_lda_mat_path):
    return (
        f'ark,s,cs:splice-feats --left-context=3 --right-context=3 scp,s,cs:"{reference_final_features_path}" ark:- '
        f'| transform-feats "{sat_lda_mat_path}" ark:- ark:- |'
    )


@pytest.fixture(scope="session")
def reference_sat_feature_string(
    reference_dir, reference_final_features_path, sat_lda_mat_path, reference_trans_path
):
    utt2spk = reference_dir.joinpath("utt2spk.scp")
    return (
        f'ark,s,cs:splice-feats --left-context=3 --right-context=3 scp,s,cs:"{reference_final_features_path}" ark:- '
        f'| transform-feats "{sat_lda_mat_path}" ark:- ark:- | transform-feats --utt2spk=ark:"{utt2spk}" scp:"{reference_trans_path}" ark:- ark:- |'
    )


@pytest.fixture(scope="session")
def reference_first_pass_ali_path(
    wav_path,
    reference_dir,
    sat_align_model_path,
    sat_temp_dir,
    align_options,
    reference_si_feature_string,
):
    ali_path = reference_dir.joinpath("ali_first_pass.ark")
    fst_path = sat_temp_dir.joinpath("fsts.ark")
    subprocess.call(
        [
            "gmm-align-compiled",
            f"--transition-scale={align_options['transition_scale']}",
            f"--acoustic-scale={align_options['acoustic_scale']}",
            f"--self-loop-scale={align_options['self_loop_scale']}",
            f"--beam={align_options['beam']}",
            f"--retry-beam={align_options['retry_beam']}",
            "--careful=false",
            sat_align_model_path,
            f"ark,s,cs:{fst_path}",
            reference_si_feature_string,
            f"ark:{ali_path}",
        ],
        env=os.environ,
    )
    return ali_path


@pytest.fixture(scope="session")
def reference_second_pass_ali_path(
    wav_path,
    reference_dir,
    sat_model_path,
    sat_temp_dir,
    align_options,
    reference_sat_feature_string,
):
    ali_path = reference_dir.joinpath("ali_second_pass.ark")
    fst_path = sat_temp_dir.joinpath("fsts.ark")
    subprocess.call(
        [
            "gmm-align-compiled",
            f"--transition-scale={align_options['transition_scale']}",
            f"--acoustic-scale={align_options['acoustic_scale']}",
            f"--self-loop-scale={align_options['self_loop_scale']}",
            f"--beam={align_options['beam']}",
            f"--retry-beam={align_options['retry_beam']}",
            "--careful=false",
            sat_model_path,
            f"ark,s,cs:{fst_path}",
            reference_sat_feature_string,
            f"ark:{ali_path}",
        ],
        env=os.environ,
    )
    return ali_path


@pytest.fixture(scope="session")
def align_options():
    return {
        "transition_scale": 1.0,
        "acoustic_scale": 0.1,
        "self_loop_scale": 0.1,
        "beam": 10,
        "retry_beam": 40,
    }


@pytest.fixture(scope="session")
def fmllr_options():
    return {"fmllr_update_type": "full", "silence_weight": 0.0, "silence_csl": "1:2"}


@pytest.fixture(scope="session")
def reference_trans_path(
    wav_path,
    reference_dir,
    reference_final_features_path,
    reference_si_feature_string,
    fmllr_options,
    sat_align_model_path,
    sat_model_path,
):
    ark_path = reference_dir.joinpath("trans.ark")
    scp_path = reference_dir.joinpath("trans.scp")
    ali_path = reference_dir.joinpath("ali_first_pass.ark")
    spk2utt = reference_dir.joinpath("spk2utt.scp")

    post_proc = subprocess.Popen(
        ["ali-to-post", f"ark,s,cs:{ali_path}", "ark:-"],
        stdout=subprocess.PIPE,
        env=os.environ,
    )

    weight_proc = subprocess.Popen(
        [
            "weight-silence-post",
            "0.0",
            fmllr_options["silence_csl"],
            sat_align_model_path,
            "ark,s,cs:-",
            "ark:-",
        ],
        stdin=post_proc.stdout,
        stdout=subprocess.PIPE,
        env=os.environ,
    )
    post_gpost_proc = subprocess.Popen(
        [
            "gmm-post-to-gpost",
            sat_align_model_path,
            reference_si_feature_string,
            "ark,s,cs:-",
            "ark:-",
        ],
        stdin=weight_proc.stdout,
        stdout=subprocess.PIPE,
        env=os.environ,
    )
    est_proc = subprocess.Popen(
        [
            "gmm-est-fmllr-gpost",
            f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
            f"--spk2utt=ark:{spk2utt}",
            sat_model_path,
            reference_si_feature_string,
            "ark,s,cs:-",
            f"ark,scp:{ark_path},{scp_path}",
        ],
        encoding="utf8",
        stdin=post_gpost_proc.stdout,
        env=os.environ,
    )
    est_proc.communicate()
    return scp_path


@pytest.fixture(scope="session")
def reference_trans_compose_path(
    wav_path,
    reference_dir,
    reference_final_features_path,
    reference_sat_feature_string,
    fmllr_options,
    sat_model_path,
    reference_trans_path,
    reference_second_pass_ali_path,
):
    temp_ark_path = reference_dir.joinpath("trans_second.ark")
    temp_scp_path = reference_dir.joinpath("trans_second.scp")
    ark_path = reference_dir.joinpath("trans_composed.ark")
    scp_path = reference_dir.joinpath("trans_composed.scp")
    spk2utt = reference_dir.joinpath("spk2utt.scp")

    post_proc = subprocess.Popen(
        ["ali-to-post", f"ark,s,cs:{reference_second_pass_ali_path}", "ark:-"],
        stdout=subprocess.PIPE,
        env=os.environ,
    )

    weight_proc = subprocess.Popen(
        [
            "weight-silence-post",
            "0.0",
            fmllr_options["silence_csl"],
            sat_model_path,
            "ark,s,cs:-",
            "ark:-",
        ],
        stdin=post_proc.stdout,
        stdout=subprocess.PIPE,
        env=os.environ,
    )
    est_proc = subprocess.Popen(
        [
            "gmm-est-fmllr",
            f"--fmllr-update-type={fmllr_options['fmllr_update_type']}",
            f"--spk2utt=ark:{spk2utt}",
            sat_model_path,
            reference_sat_feature_string,
            "ark,s,cs:-",
            f"ark,scp:{temp_ark_path},{temp_scp_path}",
        ],
        encoding="utf8",
        stdin=weight_proc.stdout,
        env=os.environ,
    )
    est_proc.communicate()
    compose_proc = subprocess.Popen(
        [
            "compose-transforms",
            "--b-is-affine=true",
            f"scp:{temp_scp_path}",
            f"scp:{reference_trans_path}",
            f"ark,scp:{ark_path},{scp_path}",
        ],
        env=os.environ,
    )
    compose_proc.communicate()
    return scp_path


@pytest.fixture(scope="session")
def mono_temp_dir(temp_dir):
    p = temp_dir.joinpath("mono")
    p.mkdir(exist_ok=True)
    return p


@pytest.fixture(scope="session")
def sat_temp_dir(temp_dir):
    p = temp_dir.joinpath("sat")
    p.mkdir(exist_ok=True)
    return p


@pytest.fixture(scope="session")
def dictionaries_dir(test_dir):
    return test_dir.joinpath("dictionaries")


@pytest.fixture(scope="session")
def lm_dir(test_dir):
    return test_dir.joinpath("language_models")


@pytest.fixture(scope="session")
def acoustic_corpus_text():
    return "this is the acoustic corpus i'm talking pretty fast here there's nothing going else going on we're just yknow there's some speech errors but who cares um this is me talking really slow and slightly lower in intensity uh we're just saying some words and here's some more words words words words um and that should be all thanks"


@pytest.fixture(scope="session")
def sat_phones():
    return [
        "a",
        "aj",
        "aw",
        "aː",
        "b",
        "bʲ",
        "c",
        "cʰ",
        "cʷ",
        "d",
        "dʒ",
        "dʲ",
        "d̪",
        "e",
        "ej",
        "eː",
        "f",
        "fʲ",
        "fʷ",
        "h",
        "i",
        "iː",
        "j",
        "k",
        "kp",
        "kʰ",
        "kʷ",
        "l",
        "m",
        "mʲ",
        "m̩",
        "n",
        "n̩",
        "o",
        "ow",
        "oː",
        "p",
        "pʰ",
        "pʲ",
        "pʷ",
        "s",
        "t",
        "tʃ",
        "tʰ",
        "tʲ",
        "tʷ",
        "t̪",
        "u",
        "uː",
        "v",
        "vʲ",
        "vʷ",
        "w",
        "z",
        "æ",
        "ç",
        "ð",
        "ŋ",
        "ɐ",
        "ɑ",
        "ɑː",
        "ɒ",
        "ɒː",
        "ɔ",
        "ɔj",
        "ɖ",
        "ə",
        "əw",
        "ɚ",
        "ɛ",
        "ɛː",
        "ɜ",
        "ɜː",
        "ɝ",
        "ɟ",
        "ɟʷ",
        "ɡ",
        "ɡb",
        "ɡʷ",
        "ɪ",
        "ɫ",
        "ɫ̩",
        "ɱ",
        "ɲ",
        "ɹ",
        "ɾ",
        "ɾʲ",
        "ɾ̃",
        "ʃ",
        "ʈ",
        "ʈʲ",
        "ʈʷ",
        "ʉ",
        "ʉː",
        "ʊ",
        "ʋ",
        "ʎ",
        "ʒ",
        "ʔ",
        "θ",
    ]


@pytest.fixture(scope="session")
def wav_path(wav_dir):
    return wav_dir.joinpath("acoustic_corpus.wav")


@pytest.fixture(scope="session")
def mfa_wav_path(wav_dir):
    return wav_dir.joinpath("mfa_michael.flac")


@pytest.fixture(scope="session")
def lm_path(lm_dir):
    return lm_dir.joinpath("test_lm.arpa")


@pytest.fixture(scope="session")
def mono_model_path(mono_am_dir):
    return mono_am_dir.joinpath("final.mdl")


@pytest.fixture(scope="session")
def mono_tree_path(mono_am_dir):
    return mono_am_dir.joinpath("tree")


@pytest.fixture(scope="session")
def sat_align_model_path(sat_am_dir):
    return sat_am_dir.joinpath("final.alimdl")


@pytest.fixture(scope="session")
def sat_model_path(sat_am_dir):
    return sat_am_dir.joinpath("final.mdl")


@pytest.fixture(scope="session")
def sat_tree_path(sat_am_dir):
    return sat_am_dir.joinpath("tree")


@pytest.fixture(scope="session")
def sat_lda_mat_path(sat_am_dir):
    return sat_am_dir.joinpath("lda.mat")


@pytest.fixture(scope="session")
def dictionary_path(dictionaries_dir):
    return dictionaries_dir.joinpath("test_basic.txt")


@pytest.fixture(scope="session")
def sat_dictionary_path(dictionaries_dir):
    return dictionaries_dir.joinpath("test_sat.txt")


@pytest.fixture(scope="session")
def mfa_dictionary_path(dictionaries_dir):
    return dictionaries_dir.joinpath("test_mfa.txt")


@pytest.fixture(scope="session")
def reference_hello_intervals():
    from _kalpy.util import Interval

    return [
        Interval(0.0, 1.0, "hh"),
        Interval(1.0, 2.0, "eh"),
        Interval(2.0, 3.0, "l"),
        Interval(3.0, 4.0, "ow"),
    ]


@pytest.fixture(scope="session")
def test_hi_intervals():
    from _kalpy.util import Interval

    return [
        Interval(0.0, 1.0, "hh"),
        Interval(1.0, 2.0, "ay"),
    ]



"""Classes for generating training graphs"""
import logging
import os
import pathlib
import re
import typing

import pywrapfst

from _kalpy.decoder import TrainingGraphCompilerOptions
from _kalpy.fstext import (
    ConstFst,
    VectorFst,
    fst_add_self_loops,
    fst_arc_sort,
    fst_compose_context,
    fst_determinize_star,
    fst_minimize_encoded,
    fst_push_special,
    fst_rm_eps_local,
    fst_rm_symbols,
    fst_table_compose,
)
from _kalpy.hmm import TransitionModel, make_h_transducer
from _kalpy.lm import ArpaParseOptions, BuildConstArpaLm, ConstArpaLm, arpa_to_fst
from _kalpy.tree import ContextDependency
from _kalpy.util import ReadKaldiObject
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.fstext.utils import kaldi_to_pynini, pynini_to_kaldi, pynini_to_kaldi_const

logger = logging.getLogger("kalpy.decode_graph")
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class DecodeGraphCompiler:
    """
    Parameters
    ----------
    acoustic_model_path: str
        Path to model file
    tree_path: str
        Path to tree file
    lexicon_compiler: :class:`~kalpy.fstext.lexicon.LexiconCompiler`
        Lexicon compiler to use in generating training graphs
    transition_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment
    self_loop_scale: float
        Scale on transitions, typically set to 0 as it will be defined during alignment

    Attributes
    ----------
    transition_model: :class:`~_kalpy.hmm.TransitionModel`
        Transition model to use in compiling the graphs
    tree: :class:`~_kalpy.hmm.TransitionModel`
        Context dependency tree to use in compiling the graphs
    """

    def __init__(
        self,
        acoustic_model_path: typing.Union[str, pathlib.Path],
        tree_path: typing.Union[str, pathlib.Path],
        lexicon_compiler: LexiconCompiler,
        arpa_path: typing.Union[str, pathlib.Path] = None,
        transition_scale: float = 1.0,
        self_loop_scale: float = 0.1,
    ):
        self.acoustic_model_path = acoustic_model_path
        self.transition_model = TransitionModel()
        ReadKaldiObject(str(acoustic_model_path), self.transition_model)
        self.tree = ContextDependency()
        ReadKaldiObject(str(tree_path), self.tree)
        self.lexicon_compiler = lexicon_compiler
        self.context_width = self.tree.ContextWidth()
        self.central_pos = self.tree.CentralPosition()
        self.options = TrainingGraphCompilerOptions(transition_scale, self_loop_scale)
        self.hclg_fst = None
        self.g_fst = None
        self.lg_fst = None
        self.g_carpa_path = None
        if arpa_path is not None:
            self.compile_hclg_fst(arpa_path)

    def export_hclg(
        self,
        arpa_path: typing.Union[pathlib.Path, str],
        file_name: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Export HCLG.fst

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Path to save HCLG.fst
        arpa_path: :class:`~pathlib.Path` or str
            Path to ARPA format language model
        """
        file_name = str(file_name)
        if self.hclg_fst is None:
            self.compile_hclg_fst(str(arpa_path))
        self.hclg_fst.Write(file_name)

    def export_g(
        self,
        file_name: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Export g.fst

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Path to save g.fst
        """
        if self.g_fst is None:
            return
        file_name = str(file_name)
        self.g_fst.Write(file_name)

    def load_from_file(
        self,
        hclg_fst_path: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Read HCLG.fst from file

        Parameters
        ----------
        hclg_fst_path: :class:`~pathlib.Path` or str
            Path to read HCLG.fst
        """
        hclg_fst_path = str(hclg_fst_path)
        self.hclg_fst = ConstFst.Read(hclg_fst_path)

    def load_g_from_file(
        self,
        g_fst_path: typing.Union[pathlib.Path, str],
    ) -> None:
        """
        Read g.fst from file

        Parameters
        ----------
        g_fst_path: :class:`~pathlib.Path` or str
            Path to read HCLG.fst
        """
        g_fst_path = str(g_fst_path)
        self.g_fst = VectorFst.Read(g_fst_path)

    def compile_g_fst(self, arpa_path: typing.Union[str, pathlib.Path]) -> VectorFst:
        """
        Compile G.fst for a language model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA format language model

        Returns
        -------
        :class:`_kalpy.fstext.VectorFst`
            G.fst
        """
        if self.g_fst is None:
            self.g_fst = arpa_to_fst(str(arpa_path), self.lexicon_compiler.word_table)

        return self.g_fst

    @property
    def g_carpa(self) -> ConstArpaLm:
        g_carpa = ConstArpaLm()
        ReadKaldiObject(self.g_carpa_path, g_carpa)
        return g_carpa

    def compile_g_carpa(
        self,
        arpa_path: typing.Union[str, pathlib.Path],
        compiled_path: typing.Union[str, pathlib.Path],
    ) -> None:
        """
        Compile G.carpa for a language model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA format language model
        compiled_path: str
            Path to compiled G.carpa
        """
        compiled_path = str(compiled_path)
        temp_carpa_path = compiled_path + ".temp"
        with open(arpa_path, "r", encoding="utf8") as f, open(
            temp_carpa_path, "w", encoding="utf8"
        ) as outf:
            current_order = -1
            num_oov_lines = 0
            for line in f:
                line = line.strip()
                col = line.split()
                if current_order == -1 and not re.match(r"^\\data\\$", line):
                    continue
                if re.match(r"^\\data\\$", line):
                    current_order = 0
                    outf.write(line + "\n")
                elif re.match(r"^\\[0-9]*-grams:$", line):
                    current_order = int(re.sub(r"\\([0-9]*)-grams:$", r"\1", line))
                    outf.write(line + "\n")
                elif re.match(r"^\\end\\$", line):
                    outf.write(line + "\n")
                elif not line:
                    if current_order >= 1:
                        outf.write("\n")
                else:
                    if current_order == 0:
                        outf.write(line + "\n")
                    else:
                        if len(col) > 2 + current_order or len(col) < 1 + current_order:
                            raise Exception(f'Bad line in arpa lm "{line}"')
                        prob = col.pop(0)
                        is_oov = False
                        for i in range(current_order):
                            if not self.lexicon_compiler.word_table.member(col[i]):
                                is_oov = True
                                num_oov_lines += 1
                                break
                            col[i] = str(self.lexicon_compiler.word_table.find(col[i]))
                        if not is_oov:
                            rest_of_line = " ".join(col)
                            outf.write(f"{prob}\t{rest_of_line}\n")
        options = ArpaParseOptions()
        options.bos_symbol = self.lexicon_compiler.word_table.find("<s>")
        options.eos_symbol = self.lexicon_compiler.word_table.find("</s>")
        options.unk_symbol = self.lexicon_compiler.word_table.find(self.lexicon_compiler.oov_word)
        self.g_carpa_path = str(compiled_path)
        success = BuildConstArpaLm(options, str(temp_carpa_path), self.g_carpa_path)
        if not success:
            logger.error(f"Error compiling G.carpa from {arpa_path}")
        else:
            os.remove(temp_carpa_path)

    def compile_lg_fst(self, arpa_path: typing.Union[str, pathlib.Path]) -> VectorFst:
        """
        Compile LG.fst for a language model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA format language model

        Returns
        -------
        :class:`_kalpy.fstext.VectorFst`
            LG.fst
        """
        g_fst = self.compile_g_fst(arpa_path)
        l_fst = pynini_to_kaldi(self.lexicon_compiler.fst)
        lg_fst = fst_table_compose(l_fst, g_fst)
        fst_determinize_star(lg_fst, use_log=True)
        fst_minimize_encoded(lg_fst)
        fst_push_special(lg_fst)
        self.lg_fst = lg_fst
        return lg_fst

    def compile_clg_fst(
        self, arpa_path: typing.Union[str, pathlib.Path]
    ) -> typing.Tuple[VectorFst, typing.List[typing.List[int]]]:
        """
        Compile CLG.fst for a language model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA format language model

        Returns
        -------
        :class:`_kalpy.fstext.VectorFst`
            CLG.fst
        list[list[int]]
            Ilabels in CLG FST
        """
        if self.lg_fst is None:
            self.lg_fst = self.compile_lg_fst(arpa_path)
        clg_fst, disambig_out, ilabels = fst_compose_context(
            self.lg_fst,
            self.lexicon_compiler.disambiguation_symbols,
            self.context_width,
            self.central_pos,
        )

        fst_arc_sort(clg_fst, sort_type="ilabel")
        return clg_fst, ilabels

    def compile_hclg_fst(self, arpa_path: typing.Union[str, pathlib.Path]):
        """
        Compile HCLG.fst for a language model

        Parameters
        ----------
        arpa_path: str
            Path to ARPA format language model

        Returns
        -------
        :class:`_kalpy.fstext.ConstFst`
            HCLG.fst
        """
        clg_fst, ilabels = self.compile_clg_fst(arpa_path)

        h, disambig = make_h_transducer(self.tree, self.transition_model, ilabels)
        hclga_fst = fst_table_compose(h, clg_fst)
        fst_determinize_star(hclga_fst, use_log=True)
        fst_rm_symbols(hclga_fst, disambig)
        fst_rm_eps_local(hclga_fst)
        fst_minimize_encoded(hclga_fst)
        fst_add_self_loops(hclga_fst, self.transition_model, [], self.options.self_loop_scale)
        hclg = pywrapfst.convert(kaldi_to_pynini(hclga_fst), "const")
        self.hclg_fst = pynini_to_kaldi_const(hclg)



import pynini
import pywrapfst

from _kalpy.fstext import VectorFst


def test_fst_casting():
    fst = pynini.Fst()
    start = fst.add_state()
    end = fst.add_state()
    fst.set_start(start)
    fst.set_final(end, 0)
    fst.add_arc(start, pywrapfst.Arc(0, 0, 0, end))
    kaldi_fst = VectorFst.from_pynini(fst)
    assert kaldi_fst.Start() == start
    assert kaldi_fst.NumStates() == 2
    assert kaldi_fst.NumArcs(start) == 1


def test_fst_read(mono_temp_dir):
    fst_path = mono_temp_dir.joinpath("lexicon.fst")
    kaldi_fst = VectorFst.Read(str(fst_path))
    assert kaldi_fst.Start() == 0
    assert kaldi_fst.NumStates() > 0



"""
Model classes
=============

"""
from __future__ import annotations

import json
import logging
import os
import typing
from pathlib import Path

import pywrapfst
import yaml

from _kalpy.feat import (
    DeltaFeaturesOptions,
    SlidingWindowCmnOptions,
    compute_deltas,
    paste_feats,
    sliding_window_cmn,
    splice_frames,
)
from _kalpy.matrix import DoubleMatrix, FloatMatrix
from _kalpy.transform import ApplyCmvn, apply_transform
from kalpy.data import Segment
from kalpy.exceptions import AcousticModelError
from kalpy.feat.cmvn import CmvnComputer
from kalpy.feat.mfcc import MfccComputer
from kalpy.feat.pitch import PitchComputer
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.gmm.utils import read_gmm_model
from kalpy.utils import read_kaldi_object

if typing.TYPE_CHECKING:
    from _kalpy.gmm import AmDiagGmm
    from _kalpy.hmm import TransitionModel

logger = logging.getLogger("mfa")


class AcousticModel:
    """
    Class for storing acoustic models in MFA, exported as zip files containing the necessary Kaldi files
    to be reused

    """

    def __init__(
        self,
        directory: typing.Union[str, Path],
        validate: bool = True,
    ):
        self.directory = Path(directory)
        if validate:
            self.validate_directory()
        self._am = None
        self._tm = None
        self._meta = None
        self._phone_pdf_counts = None

    def validate_directory(self):
        required_files = [
            "final.mdl",
            "tree",
        ]
        for f in required_files:
            if not self.directory.joinpath(f).exists():
                raise AcousticModelError(f"Could not find {f} in {self.directory}.")

    def generate_features(
        self,
        segment: Segment,
        cmvn: typing.Optional[DoubleMatrix] = None,
        fmllr_trans: FloatMatrix = None,
        uses_speaker_adaptation: bool = True,
        uses_splices: bool = False,
        uses_deltas: bool = True,
        splice_context: int = 3,
    ):
        if self.lda_mat is not None:
            uses_splices = True
            uses_deltas = False
        feats = self.mfcc_computer.compute_mfccs_for_export(segment, compress=False)
        if cmvn is None:
            cmvn_computer = CmvnComputer()
            cmvn = cmvn_computer.compute_cmvn_from_features([feats])
        ApplyCmvn(cmvn, False, feats)

        if self.pitch_computer is not None:
            pitch = self.pitch_computer.compute_pitch_for_export(segment, compress=False)
            feats = paste_feats([feats, pitch], 1)
        if uses_splices:
            feats = splice_frames(feats, splice_context, splice_context)
            if self.lda_mat is not None:
                feats = apply_transform(feats, self.lda_mat)
        elif uses_deltas:
            delta_options = DeltaFeaturesOptions()
            feats = compute_deltas(delta_options, feats)
        if uses_speaker_adaptation and fmllr_trans is not None:
            feats = apply_transform(feats, fmllr_trans)
        return feats

    def generate_features_for_fine_tune(
        self,
        segment: Segment,
        cmvn: typing.Optional[DoubleMatrix] = None,
        fmllr_trans: FloatMatrix = None,
        uses_speaker_adaptation: bool = True,
        uses_splices: bool = False,
        uses_deltas: bool = True,
        splice_context: int = 3,
    ):

        mfcc_options = self.mfcc_options
        mfcc_options["frame_shift"] /= 10
        mfcc_computer = MfccComputer(**mfcc_options)
        if self.lda_mat is not None:
            uses_splices = True
            uses_deltas = False
        feats = mfcc_computer.compute_mfccs_for_export(segment, compress=False)
        if cmvn is None:
            cmvn_computer = CmvnComputer()
            cmvn = cmvn_computer.compute_cmvn_from_features([feats])
        ApplyCmvn(cmvn, False, feats)

        if self.pitch_computer is not None:
            pitch_options = self.pitch_options
            pitch_options["frame_shift"] /= 10
            pitch_computer = PitchComputer(**pitch_options)
            pitch = pitch_computer.compute_pitch_for_export(segment, compress=False)
            feats = paste_feats([feats, pitch], 1)
        if uses_splices:
            feats = splice_frames(feats, splice_context, splice_context)
            if self.lda_mat is not None:
                feats = apply_transform(feats, self.lda_mat)
        elif uses_deltas:
            delta_options = DeltaFeaturesOptions()
            feats = compute_deltas(delta_options, feats)
        if uses_speaker_adaptation and fmllr_trans is not None:
            feats = apply_transform(feats, fmllr_trans)
        return feats

    @property
    def version(self):
        return self.meta["version"]

    @property
    def uses_cmvn(self):
        return self.meta["features"]["uses_cmvn"]

    @property
    def parameters(self) -> typing.Dict[str, typing.Any]:
        """Parameters to pass to top-level workers"""
        params = {**self.meta["features"]}
        params["non_silence_phones"] = {x for x in self.meta["phones"]}
        params["oov_phone"] = self.meta["oov_phone"]
        params["language"] = self.meta["language"]
        params["optional_silence_phone"] = self.meta["optional_silence_phone"]
        params["phone_set_type"] = self.meta["phone_set_type"]
        params["silence_probability"] = self.meta.get("silence_probability", 0.5)
        params["initial_silence_probability"] = self.meta.get("initial_silence_probability", 0.5)
        params["final_non_silence_correction"] = self.meta.get(
            "final_non_silence_correction", None
        )
        params["final_silence_correction"] = self.meta.get("final_silence_correction", None)
        if "other_noise_phone" in self.meta:
            params["other_noise_phone"] = self.meta["other_noise_phone"]
        if (
            "dictionaries" in self.meta
            and "position_dependent_phones" in self.meta["dictionaries"]
        ):
            params["position_dependent_phones"] = self.meta["dictionaries"][
                "position_dependent_phones"
            ]
        else:
            params["position_dependent_phones"] = self.meta.get("position_dependent_phones", True)
        return params

    @property
    def tree_path(self) -> Path:
        """Current acoustic model path"""
        return self.directory.joinpath("tree")

    @property
    def lda_mat_path(self) -> Path:
        """Current acoustic model path"""
        return self.directory.joinpath("lda.mat")

    @property
    def model_path(self) -> Path:
        """Current acoustic model path"""
        return self.directory.joinpath("final.mdl")

    @property
    def phone_symbol_path(self) -> Path:
        """Path to phone symbol table"""
        return self.directory.joinpath("phones.txt")

    @property
    def phone_pdf_counts_path(self) -> Path:
        """Path to phone symbol table"""
        return self.directory.joinpath("phone_pdf.counts")

    @property
    def phone_pdf_counts(self):
        if not self.phone_pdf_counts_path.exists():
            return {}
        if self._phone_pdf_counts is None:
            with open(self.phone_pdf_counts_path, "r", encoding="utf8") as f:
                data = json.load(f)
            self._phone_pdf_counts = {}
            for phone, pdf_counts in data.items():
                self._phone_pdf_counts[phone] = {}
                for pdf, count in pdf_counts.items():
                    self._phone_pdf_counts[phone][int(pdf)] = count

            for phone, pdf_counts in self._phone_pdf_counts.items():
                phone_total = sum(pdf_counts.values())
                for pdf, count in pdf_counts.items():
                    self._phone_pdf_counts[phone][int(pdf)] = count / phone_total
        return self._phone_pdf_counts

    @property
    def alignment_model_path(self) -> Path:
        """Alignment model path"""
        path = self.model_path.with_suffix(".alimdl")
        if os.path.exists(path):
            return path
        return self.model_path

    @property
    def acoustic_model(self) -> AmDiagGmm:
        if not self.alignment_model_path.exists():
            raise AcousticModelError(f"Could not find {self.alignment_model_path}")
        if self._am is None:
            self._tm, self._am = read_gmm_model(self.alignment_model_path)
        return self._am

    @property
    def transition_model(self) -> TransitionModel:
        if not self.alignment_model_path.exists():
            raise AcousticModelError(f"Could not find {self.alignment_model_path}")
        if self._tm is None:
            self._tm, self._am = read_gmm_model(self.alignment_model_path)
        return self._tm

    @property
    def lexicon_compiler(self):
        lc = LexiconCompiler(
            silence_probability=self.meta.get("silence_probability", 0.5),
            initial_silence_probability=self.meta.get("initial_silence_probability", 0.5),
            final_silence_correction=self.meta.get("final_silence_correction", None),
            final_non_silence_correction=self.meta.get("final_non_silence_correction", None),
            silence_phone=self.meta.get("optional_silence_phone", "sil"),
            oov_phone=self.meta.get("oov_phone", "sil"),
            position_dependent_phones=self.meta.get("position_dependent_phones", False),
            phones={x for x in self.meta["phones"]},
        )
        if self.meta.get("phone_mapping", None):
            lc.phone_table = pywrapfst.SymbolTable()
            for k, v in self.meta["phone_mapping"].items():
                lc.phone_table.add_symbol(k, v)
        elif self.phone_symbol_path.exists():
            lc.phone_table = pywrapfst.SymbolTable.read_text(self.phone_symbol_path)
        return lc

    @property
    def mfcc_computer(self) -> MfccComputer:
        return MfccComputer(**self.mfcc_options)

    @property
    def pitch_computer(self) -> typing.Optional[PitchComputer]:
        if self.meta["features"].get("use_pitch", False):
            return PitchComputer(**self.pitch_options)
        return

    @property
    def lda_mat(self) -> FloatMatrix:
        lda_mat_path = self.directory.joinpath("lda.mat")
        lda_mat = None
        if lda_mat_path.exists():
            lda_mat = read_kaldi_object(FloatMatrix, lda_mat_path)
        return lda_mat

    @property
    def mfcc_options(self) -> typing.Dict[str, typing.Any]:
        """Parameters to use in computing MFCC features."""
        return {
            "sample_frequency": self.meta["features"].get("sample_frequency", 16000),
            "frame_shift": self.meta["features"].get("frame_shift", 10),
            "frame_length": self.meta["features"].get("frame_length", 25),
            "dither": self.meta["features"].get("dither", 0.0001),
            "preemphasis_coefficient": self.meta["features"].get("preemphasis_coefficient", 0.97),
            "snip_edges": self.meta["features"].get("snip_edges", True),
            "num_mel_bins": self.meta["features"].get("num_mel_bins", 23),
            "low_frequency": self.meta["features"].get("low_frequency", 20),
            "high_frequency": self.meta["features"].get("high_frequency", 7800),
            "num_coefficients": self.meta["features"].get("num_coefficients", 13),
            "use_energy": self.meta["features"].get("use_energy", False),
            "energy_floor": self.meta["features"].get("energy_floor", 1.0),
            "raw_energy": self.meta["features"].get("raw_energy", True),
            "cepstral_lifter": self.meta["features"].get("cepstral_lifter", 22),
        }

    @property
    def frame_shift_seconds(self):
        return round(self.meta["features"].get("frame_shift", 10) / 1000, 4)

    @property
    def pitch_options(self) -> typing.Dict[str, typing.Any]:
        """Parameters to use in computing pitch features."""
        use_pitch = self.meta["features"].get("use_pitch", False)
        use_voicing = self.meta["features"].get("use_voicing", False)
        use_delta_pitch = self.meta["features"].get("use_delta_pitch", False)
        normalize = self.meta["features"].get("normalize_pitch", True)
        options = {
            "frame_shift": self.meta["features"].get("frame_shift", 10),
            "frame_length": self.meta["features"].get("frame_length", 25),
            "min_f0": self.meta["features"].get("min_f0", 50),
            "max_f0": self.meta["features"].get("max_f0", 800),
            "sample_frequency": self.meta["features"].get("sample_frequency", 16000),
            "penalty_factor": self.meta["features"].get("penalty_factor", 0.1),
            "delta_pitch": self.meta["features"].get("delta_pitch", 0.005),
            "snip_edges": self.meta["features"].get("snip_edges", True),
            "add_normalized_log_pitch": False,
            "add_delta_pitch": False,
            "add_pov_feature": False,
        }
        if use_pitch:
            options["add_normalized_log_pitch"] = normalize
            options["add_raw_log_pitch"] = not normalize
        options["add_delta_pitch"] = use_delta_pitch
        options["add_pov_feature"] = use_voicing
        return options

    @property
    def lda_options(self) -> typing.Dict[str, typing.Any]:
        """Parameters to use in computing MFCC features."""
        return {
            "splice_left_context": self.meta["features"].get("splice_left_context", 3),
            "splice_right_context": self.meta["features"].get("splice_right_context", 3),
        }

    def _load_meta_data(self):
        default_features = {
            "feature_type": "mfcc",
            "use_energy": False,
            "frame_shift": 10,
            "snip_edges": True,
            "low_frequency": 20,
            "high_frequency": 7800,
            "sample_frequency": 16000,
            "allow_downsample": True,
            "allow_upsample": True,
            "use_pitch": False,
            "use_voicing": False,
            "uses_cmvn": True,
            "uses_deltas": True,
            "uses_splices": False,
            "uses_voiced": False,
            "uses_speaker_adaptation": False,
            "silence_weight": 0.0,
            "fmllr_update_type": "full",
            "splice_left_context": 3,
            "splice_right_context": 3,
        }
        if not self._meta:
            meta_path = self.directory.joinpath("meta.json")
            file_format = "json"
            if not os.path.exists(meta_path):
                meta_path = self.directory.joinpath("meta.yaml")
                file_format = "yaml"
            if not os.path.exists(meta_path):
                self._meta = {
                    "version": "0.9.0",
                    "architecture": "gmm-hmm",
                    "features": default_features,
                }
            else:
                with open(meta_path, "r", encoding="utf8") as f:
                    if file_format == "yaml":
                        self._meta = yaml.load(f, Loader=yaml.Loader)
                    else:
                        self._meta = json.load(f)
                if self._meta["features"] == "mfcc+deltas":
                    self._meta["features"] = default_features
                    if "pitch" in self._meta["features"]:
                        self._meta["features"]["use_pitch"] = self._meta["features"].pop("pitch")
                if (
                    self._meta["features"].get("use_pitch", False)
                    and self._meta["version"] < "2.0.6"
                ):
                    self._meta["features"]["use_delta_pitch"] = True
            if "phone_type" not in self._meta:
                self._meta["phone_type"] = "triphone"
            if "optional_silence_phone" not in self._meta:
                self._meta["optional_silence_phone"] = "sil"
            if "oov_phone" not in self._meta:
                self._meta["oov_phone"] = "spn"
            if file_format == "yaml":
                self._meta["other_noise_phone"] = "sp"
            if "phone_set_type" not in self._meta:
                self._meta["phone_set_type"] = "UNKNOWN"
            if "language" not in self._meta or self._meta["version"] < "3.0":
                self._meta["language"] = "unknown"
            self._meta["phones"] = set(self._meta.get("phones", []))
            if (
                "uses_speaker_adaptation" not in self._meta["features"]
                or not self._meta["features"]["uses_speaker_adaptation"]
            ):
                self._meta["features"]["uses_speaker_adaptation"] = os.path.exists(
                    self.directory.joinpath("final.alimdl")
                )
            if self._meta["version"] in {"0.9.0", "1.0.0"}:
                self._meta["features"]["uses_speaker_adaptation"] = True
            if (
                "uses_splices" not in self._meta["features"]
                or not self._meta["features"]["uses_splices"]
            ):
                self._meta["features"]["uses_splices"] = os.path.exists(
                    self.directory.joinpath("lda.mat")
                )
                if self._meta["features"]["uses_splices"]:
                    self._meta["features"]["uses_deltas"] = False
            if (
                self._meta["features"].get("use_pitch", False)
                and "use_voicing" not in self._meta["features"]
            ):
                self._meta["features"]["use_voicing"] = True
            if (
                "dictionaries" in self._meta
                and "position_dependent_phones" not in self._meta["dictionaries"]
            ):
                if self._meta["version"] < "2.0":
                    default_value = True
                else:
                    default_value = False
                self._meta["dictionaries"]["position_dependent_phones"] = self._meta.get(
                    "position_dependent_phones", default_value
                )
        self.parse_old_features()

    @property
    def meta(self) -> typing.Dict[str, typing.Any]:
        """
        Metadata information for the acoustic model
        """
        if not self._meta:
            self._load_meta_data()
        return self._meta

    def parse_old_features(self) -> None:
        """
        Parse MFA model's features and ensure that they are up-to-date with current functionality
        """
        if "features" not in self._meta:
            return
        feature_key_remapping = {
            "type": "feature_type",
            "deltas": "uses_deltas",
            "fmllr": "uses_speaker_adaptation",
        }

        for key, new_key in feature_key_remapping.items():
            if key in self._meta["features"]:
                self._meta["features"][new_key] = self._meta["features"][key]
                del self._meta["features"][key]
        if "uses_splices" not in self._meta["features"]:  # Backwards compatibility
            self._meta["features"]["uses_splices"] = os.path.exists(
                self.directory.joinpath("lda.mat")
            )
        if "uses_speaker_adaptation" not in self._meta["features"]:
            self._meta["features"]["uses_speaker_adaptation"] = os.path.exists(
                self.directory.joinpath("final.alimdl")
            )



"""Classes for computing fMLLR transforms"""
from __future__ import annotations

import logging
import pathlib
import threading
import typing

from _kalpy import gmm, hmm, transform
from _kalpy.util import BaseFloatMatrixWriter, ConstIntegerSet
from kalpy.data import KaldiMapping, MatrixArchive
from kalpy.feat.data import FeatureArchive
from kalpy.gmm.data import AlignmentArchive, LatticeArchive
from kalpy.gmm.utils import read_gmm_model
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.fmllr")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class FmllrComputer:
    def __init__(
        self,
        alignment_acoustic_model_path: typing.Union[pathlib.Path, str],
        acoustic_model_path: typing.Union[pathlib.Path, str],
        silence_phones: typing.List[int],
        spk2utt: KaldiMapping = None,
        weight_distribute: bool = False,
        fmllr_update_type: str = "full",
        silence_weight: float = 0.0,
        acoustic_scale: float = 1.0,
        fmllr_min_count: float = 500.0,
        fmllr_num_iters: int = 40,
        thread_lock: typing.Optional[threading.Lock] = None,
    ):
        self.acoustic_model_path = acoustic_model_path
        self.alignment_acoustic_model_path = alignment_acoustic_model_path
        self.transition_model, self.acoustic_model = read_gmm_model(self.acoustic_model_path)
        self.two_models = self.alignment_acoustic_model_path != self.acoustic_model_path
        if self.two_models:
            self.alignment_transition_model, self.alignment_acoustic_model = read_gmm_model(
                self.alignment_acoustic_model_path
            )
        else:
            self.alignment_transition_model, self.alignment_acoustic_model = (
                self.transition_model,
                self.acoustic_model,
            )
        self.spk2utt = spk2utt
        self.silence_weight = silence_weight
        self.acoustic_scale = acoustic_scale
        self.silence_phones = silence_phones
        self.weight_distribute = weight_distribute
        self.fmllr_update_type = fmllr_update_type
        self.fmllr_min_count = fmllr_min_count
        self.fmllr_num_iters = fmllr_num_iters
        self.thread_lock = thread_lock
        self.callback_frequency = 100

    def compute_fmllr(
        self,
        feature_archive: FeatureArchive,
        alignment_archive: typing.Union[AlignmentArchive, LatticeArchive],
        callback: typing.Callable = None,
    ):
        fmllr_options = transform.FmllrOptions()
        fmllr_options.update_type = self.fmllr_update_type
        fmllr_options.min_count = self.fmllr_min_count
        fmllr_options.num_iters = self.fmllr_num_iters
        use_alignment = isinstance(alignment_archive, AlignmentArchive)
        rand_prune = 0.0
        tot_impr = 0.0
        tot_t = 0.0
        num_done = 0
        num_skipped = 0
        am_dim = self.acoustic_model.Dim()
        silence_set = ConstIntegerSet(self.silence_phones)
        if self.spk2utt is not None:
            for spk, utt_list in self.spk2utt.items():
                spk_stats = transform.FmllrDiagGmmAccs(am_dim)
                logger.info(f"Processing speaker {spk}...")
                for utterance_id in utt_list:
                    try:
                        alignment = alignment_archive[utterance_id]
                    except KeyError:
                        logger.info(f"Skipping {utterance_id} due to missing alignment.")
                        num_skipped += 1
                        continue
                    if use_alignment:
                        alignment = alignment_archive[utterance_id].alignment
                    try:
                        feats = feature_archive[utterance_id]
                    except KeyError:
                        logger.info(f"Skipping {utterance_id} due to missing features.")
                        num_skipped += 1
                        continue
                    if feats.NumRows() == 0:
                        logger.warning(f"Skipping {utterance_id} due to zero-length features")
                        num_skipped += 1
                        continue
                    num_done += 1
                    if use_alignment:
                        spk_stats.accumulate_from_alignment(
                            self.alignment_transition_model,
                            self.alignment_acoustic_model,
                            self.transition_model,
                            self.acoustic_model,
                            feats,
                            alignment,
                            silence_set,
                            self.silence_weight,
                            rand_prune=rand_prune,
                            distributed=self.weight_distribute,
                            two_models=self.two_models,
                        )
                    else:
                        spk_stats.accumulate_from_lattice(
                            self.alignment_transition_model,
                            self.alignment_acoustic_model,
                            self.transition_model,
                            self.acoustic_model,
                            feats,
                            alignment,
                            silence_set,
                            self.silence_weight,
                            acoustic_scale=self.acoustic_scale,
                            rand_prune=rand_prune,
                            distributed=self.weight_distribute,
                            two_models=self.two_models,
                        )
                    if callback is not None and num_done % self.callback_frequency == 0:
                        callback(self.callback_frequency)
                if self.thread_lock is not None:
                    self.thread_lock.acquire()
                trans, impr, spk_tot_t = spk_stats.compute_transform(
                    self.acoustic_model, fmllr_options
                )
                if self.thread_lock is not None:
                    self.thread_lock.release()
                if spk_tot_t:
                    logger.debug(
                        f"For speaker {spk}, auxf-impr from fMLLR is {impr/spk_tot_t}, over {spk_tot_t} frames."
                    )
                    tot_impr += impr
                    tot_t += spk_tot_t
                    yield spk, trans
                else:
                    logger.debug(f"Skipping speaker {spk} due to no data")
        else:
            for utterance_id, feats in feature_archive:

                try:
                    alignment = alignment_archive[utterance_id]
                except KeyError:
                    logger.info(f"Skipping {utterance_id} due to missing alignment.")
                    num_skipped += 1
                    continue

                if feats.NumRows() == 0:
                    logger.warning(f"Skipping {utterance_id} due to zero-length features")
                    num_skipped += 1
                    continue
                spk_stats = transform.FmllrDiagGmmAccs(am_dim)
                if use_alignment:
                    spk_stats.accumulate_from_alignment(
                        self.transition_model,
                        self.acoustic_model,
                        feats,
                        alignment,
                        silence_set,
                        self.silence_weight,
                        rand_prune=rand_prune,
                        distributed=self.weight_distribute,
                        two_models=self.two_models,
                    )
                else:
                    spk_stats.accumulate_from_lattice(
                        self.transition_model,
                        self.acoustic_model,
                        feats,
                        alignment,
                        silence_set,
                        self.silence_weight,
                        acoustic_scale=self.acoustic_scale,
                        rand_prune=rand_prune,
                        distributed=self.weight_distribute,
                        two_models=self.two_models,
                    )
                trans = transform.compute_fmllr_transform(
                    spk_stats, self.acoustic_model.Dim(), fmllr_options
                )
                num_done += 1

                yield utterance_id, trans
                if callback is not None and num_done % self.callback_frequency == 0:
                    callback(self.callback_frequency)
        if callback is not None and num_done % self.callback_frequency:
            callback(num_done % self.callback_frequency)
        logger.info(f"Done {num_done} utterances.")
        logger.info(f"Skipped {num_skipped} utterances.")
        if tot_t:
            logger.info(
                f"Overall fMLLR auxf impr per frame is {tot_impr / tot_t} over {tot_t} frames."
            )

    def export_transforms(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        alignment_archive: typing.Union[AlignmentArchive, LatticeArchive],
        previous_transform_archive: MatrixArchive = None,
        write_scp: bool = False,
        callback: typing.Callable = None,
    ):
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = BaseFloatMatrixWriter(write_specifier)
        try:
            prev_reader = None
            if previous_transform_archive is not None:
                prev_reader = previous_transform_archive.random_reader
            for speaker, trans in self.compute_fmllr(
                feature_archive, alignment_archive, callback=callback
            ):
                if previous_transform_archive is not None:
                    if prev_reader.HasKey(speaker):
                        prev_trans = prev_reader.Value(speaker)
                        new_trans = transform.compose_transforms(trans, prev_trans, True)
                        trans = new_trans
                writer.Write(str(speaker), trans)
        finally:
            writer.Close()



"""Classes for extracting ivectors"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy.gmm import DiagGmm
from _kalpy.ivector import IvectorExtractor as KaldiIvectorExtractor
from _kalpy.ivector import ivector_extract
from _kalpy.util import BaseDoubleVectorWriter
from kalpy.feat.data import FeatureArchive
from kalpy.utils import generate_write_specifier, read_kaldi_object

logger = logging.getLogger("kalpy.ivector")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class IvectorExtractor:
    def __init__(
        self,
        dubm_path: typing.Union[str, pathlib.Path],
        ivector_extractor_path: typing.Union[str, pathlib.Path],
        acoustic_weight: float = 1.0,
        max_count: float = 0.0,
        num_gselect: int = 50,
        min_post: float = 0.0,
    ):
        self.dubm_path = str(dubm_path)
        self.ivector_extractor_path = str(ivector_extractor_path)
        self.ivector_extractor = read_kaldi_object(
            KaldiIvectorExtractor, self.ivector_extractor_path
        )
        self.dubm = read_kaldi_object(DiagGmm, self.dubm_path)

        self.acoustic_weight = acoustic_weight
        self.max_count = max_count
        self.num_gselect = num_gselect
        self.min_post = min_post

    def extract_ivectors(self, feature_archive: FeatureArchive):
        num_done = 0
        num_error = 0
        for utt_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.warning(f"Skipping {utt_id} due to zero-length features")
                num_error += 1
                continue
            try:
                ivector = ivector_extract(
                    self.dubm,
                    self.ivector_extractor,
                    feats,
                    acoustic_weight=self.acoustic_weight,
                    max_count=self.max_count,
                    num_post=self.num_gselect,
                    min_post=self.min_post,
                )
            except Exception:
                num_error += 1
                continue
            yield utt_id, ivector
            num_done += 1
        logger.info(f"Done {num_done} utterances, errors on {num_error}.")

    def export_ivectors(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        write_scp: bool = False,
        callback: typing.Callable = None,
    ):
        write_specifier = generate_write_specifier(file_name, write_scp)
        writer = BaseDoubleVectorWriter(write_specifier)
        try:
            for utt_id, ivector in self.extract_ivectors(feature_archive):
                if callback:
                    callback(utt_id)
                writer.Write(str(utt_id), ivector)
        finally:
            writer.Close()



from _kalpy.hmm import TransitionModel
from _kalpy.util import ReadKaldiObject
from kalpy.utils import read_kaldi_object


def test_transition_model(mono_model_path):
    tm = TransitionModel()
    ReadKaldiObject(str(mono_model_path), tm)
    assert len(tm.GetPhones()) > 0


def test_read_kaldi_hmm(mono_model_path):
    tm = read_kaldi_object(TransitionModel, mono_model_path)
    assert len(tm.GetPhones()) > 0



"""Data classes for GMM"""
from __future__ import annotations

import os
import pathlib
import sys
import typing
from dataclasses import dataclass

import numpy as np
import pywrapfst
from praatio import textgrid as tgio
from praatio.data_classes.textgrid import _tgToDictionary
from praatio.utilities.constants import Interval as PraatInterval
from praatio.utilities.constants import TextgridFormats

from _kalpy.fstext import GetLinearSymbolSequence, VectorFst
from _kalpy.hmm import SplitToPhones, TransitionModel
from _kalpy.lat import (
    CompactLattice,
    CompactLatticeToWordAlignment,
    Lattice,
    RandomAccessCompactLatticeReader,
    RandomAccessLatticeReader,
    SequentialCompactLatticeReader,
    SequentialLatticeReader,
    WordAlignLatticeLexiconInfo,
    WordAlignLatticeLexiconOpts,
    lattice_best_path,
    linear_to_lattice,
    word_align_lattice_lexicon,
)
from _kalpy.matrix import FloatVector
from _kalpy.util import (
    CtmInterval,
    Interval,
    RandomAccessBaseFloatVectorReader,
    RandomAccessInt32VectorReader,
    SequentialBaseFloatVectorReader,
    SequentialInt32VectorReader,
    WordCtmInterval,
)
from kalpy.exceptions import CtmError
from kalpy.utils import generate_read_specifier


def to_tg_interval(
    interval: typing.Union[Interval, CtmInterval, WordCtmInterval], file_duration=None
) -> PraatInterval:
    """
    Converts the Kalpy's CtmInterval to
    `PraatIO's Interval class <http://timmahrt.github.io/praatIO/praatio/utilities/constants.html#Interval>`_

    Returns
    -------
    :class:`praatio.utilities.constants.Interval`
        Derived PraatIO Interval
    """
    if interval.end < -1 or interval.begin == 1000000:
        raise CtmError(interval)
    end = round(interval.end, 6)
    begin = round(interval.begin, 6)
    if begin < 0:
        begin = 0
    if file_duration is not None and end > file_duration:
        end = round(file_duration, 6)
    assert begin < end
    return PraatInterval(round(interval.begin, 6), end, interval.label)


@dataclass
class HierarchicalCtm:
    word_intervals: typing.List[WordCtmInterval]
    text: str = None
    likelihood: float = None

    def export_textgrid(
        self,
        file_name: typing.Union[str, pathlib.Path],
        file_duration: float = None,
        output_format: typing.Literal[
            "short_textgrid", "long_textgrid", "json", "textgrid_json"
        ] = "short_textgrid",
    ):
        if file_duration is not None:
            file_duration = round(file_duration, 6)
        tg = tgio.Textgrid()
        tg.minTimestamp = 0
        tg.maxTimestamp = file_duration

        word_tier, phone_tier = self.to_textgrid_tiers(file_duration)
        tg.addTier(word_tier)
        tg.addTier(phone_tier)
        if str(file_name) == "-":

            tgAsDict = _tgToDictionary(tg)

            textgridStr = tgio.textgrid_io.getTextgridAsStr(
                tgAsDict,
                output_format,
                True,
                0,
                file_duration,
            )
            sys.stdout.write(textgridStr)
            sys.stdout.flush()
        else:
            tg.save(
                str(file_name),
                includeBlankSpaces=True,
                format=output_format,
                reportingMode="error",
            )

    @property
    def phone_boundaries(self):
        boundaries = []
        for i, wi in enumerate(self.word_intervals):
            for j, pi in enumerate(wi.phones):
                if i + j != 0:
                    boundaries.append(pi.begin)
        return boundaries

    @property
    def labelled_midpoints(self):
        labels = []
        for wi in self.word_intervals:
            for pi in wi.phones:
                labels.append((pi.label, (pi.begin + pi.end) / 2))
        return labels

    def to_textgrid_tiers(
        self, file_duration: float = None
    ) -> typing.Tuple[tgio.IntervalTier, tgio.IntervalTier]:
        if file_duration is not None:
            file_duration = round(file_duration, 6)
        word_tier = tgio.IntervalTier("words", [], minT=0.0, maxT=file_duration)
        phone_tier = tgio.IntervalTier("phones", [], minT=0.0, maxT=file_duration)
        for w in self.word_intervals:
            word_tier.insertEntry(to_tg_interval(w, file_duration))
            for p in w.phones:
                phone_tier.insertEntry(to_tg_interval(p, file_duration))
        return word_tier, phone_tier

    def update_utterance_boundaries(self, begin, end=None):
        for w in self.word_intervals:
            for p in w.phones:
                p.begin += begin
                p.end += begin
        if self.word_intervals:
            self.word_intervals[-1].phones[-1].end = end


@dataclass
class Alignment:
    utterance_id: str
    alignment: typing.List[int]
    words: typing.List[int]
    likelihood: float = None
    per_frame_likelihoods: FloatVector = None
    lattice: Lattice = None

    @property
    def num_frames(self):
        return len(self.alignment)

    def generate_word_ctm(
        self,
        transition_model: TransitionModel,
        align_lexicon: WordAlignLatticeLexiconInfo,
        word_table: pywrapfst.SymbolTable,
        frame_shift: float = 0.01,
    ):
        lattice = linear_to_lattice(self.alignment, self.words)
        success, lattice = word_align_lattice_lexicon(
            lattice, transition_model, align_lexicon, WordAlignLatticeLexiconOpts()
        )
        if not success:
            raise Exception("Error aligning lattice")
        success, words, times, lengths = CompactLatticeToWordAlignment(lattice)
        if not success:
            raise Exception("Error generating word CTM")
        ctm_output = []
        for w, time, length in zip(words, times, lengths):
            word_start = round(time * frame_shift, 3)
            word_end = round((time * frame_shift) + (length * frame_shift), 3)
            label = word_table.find(w)
            ctm_output.append(CtmInterval(word_start, word_end, label, w))
        return ctm_output

    def generate_ctm(
        self,
        transition_model: TransitionModel,
        phone_table: pywrapfst.SymbolTable,
        frame_shift: float = 0.01,
    ):
        split = SplitToPhones(transition_model, self.alignment)
        ctm_output = []
        phone_start = 0.0
        current_phone_index = 0
        likelihoods = None
        if self.per_frame_likelihoods:
            likelihoods = self.per_frame_likelihoods.numpy()
        for i, s in enumerate(split):
            phone_id = transition_model.TransitionIdToPhone(s[0])
            num_repeats = len(s)
            duration = frame_shift * num_repeats
            phone_end = phone_start + duration
            label = phone_table.find(phone_id)
            confidence = 0.0
            if likelihoods is not None:
                confidence = float(
                    np.mean(likelihoods[current_phone_index : current_phone_index + num_repeats])
                )
            ctm_output.append(
                CtmInterval(
                    round(phone_start, 3),
                    round(phone_end, 3),
                    label,
                    phone_id,
                    confidence,
                )
            )
            phone_start += duration
            current_phone_index += num_repeats

        return ctm_output

    def export_textgrid(
        self,
        file_name: str,
        transition_model: TransitionModel,
        align_lexicon: VectorFst,
        word_table: pywrapfst.SymbolTable,
        phone_table: pywrapfst.SymbolTable,
        frame_shift: float = 0.01,
        duration: float = None,
        output_format: str = TextgridFormats.LONG_TEXTGRID,
    ):
        # Create initial textgrid
        tg = tgio.Textgrid()
        tg.minTimestamp = 0
        tg.maxTimestamp = duration

        intervals = self.generate_ctm(transition_model, phone_table, frame_shift=frame_shift)
        word_intervals = self.generate_word_ctm(
            transition_model, align_lexicon, word_table, frame_shift=frame_shift
        )
        if "words" not in tg.tierNames:
            tg.addTier(tgio.IntervalTier("words", [], minT=0, maxT=duration))
        if "phones" not in tg.tierNames:
            tg.addTier(tgio.IntervalTier("phones", [], minT=0, maxT=duration))
        phone_tier = tg.getTier("phones")
        word_tier = tg.getTier("words")
        for i, a in enumerate(sorted(intervals, key=lambda x: x.begin)):
            if i == len(intervals) - 1:
                a.end = duration
            if i > 0 and phone_tier.entries[-1].end > to_tg_interval(a).start:
                a.begin = phone_tier.entries[-1].end
            phone_tier.insertEntry(to_tg_interval(a, duration))
        for i, a in enumerate(sorted(word_intervals, key=lambda x: x.begin)):
            if i == len(word_intervals) - 1:
                a.end = duration
            if i > 0 and word_tier.entries[-1].end > to_tg_interval(a).start:
                a.begin = word_tier.entries[-1].end
            word_tier.insertEntry(to_tg_interval(a, duration))
        for tier in tg.tiers:
            if len(tier.entries) > 0 and tier.entries[-1][1] > tg.maxTimestamp:
                tier.insertEntry(
                    Interval(tier.entries[-1].start, tg.maxTimestamp, tier.entries[-1].label),
                    collisionMode="replace",
                )
        tg.save(
            str(file_name),
            includeBlankSpaces=True,
            format=output_format,
            reportingMode="error",
        )


class AlignmentArchive:
    """
    Class for reading an archive or SCP of alignments

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(
        self,
        file_name: typing.Union[pathlib.Path, str],
        words_file_name: typing.Union[pathlib.Path, str] = None,
        likelihood_file_name: typing.Union[pathlib.Path, str] = None,
    ):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.read_specifier = generate_read_specifier(file_name)

        self.words_reader = None
        self.words_file_name = words_file_name
        if words_file_name:
            words_read_specifier = generate_read_specifier(words_file_name)
            self.words_reader = RandomAccessInt32VectorReader(words_read_specifier)

        self.likelihood_reader = None
        self.likelihood_file_name = likelihood_file_name
        if likelihood_file_name:
            likelihood_read_specifier = generate_read_specifier(likelihood_file_name)
            self.likelihood_reader = RandomAccessBaseFloatVectorReader(likelihood_read_specifier)

        self.random_reader = RandomAccessInt32VectorReader(self.read_specifier)

    def __del__(self):
        self.close()

    def close(self):
        if self.random_reader.IsOpen():
            self.random_reader.Close()
        if self.words_reader is not None and self.words_reader.IsOpen():
            self.words_reader.Close()
        if self.likelihood_reader is not None and self.likelihood_reader.IsOpen():
            self.likelihood_reader.Close()

    @property
    def sequential_reader(self) -> SequentialInt32VectorReader:
        """Sequential reader for alignments"""
        return SequentialInt32VectorReader(self.read_specifier)

    def __iter__(self) -> typing.Generator[Alignment]:
        """Iterate over the utterance alignments in the archive"""
        reader = self.sequential_reader
        words_reader = None
        likelihood_reader = None
        if self.words_file_name is not None:
            words_read_specifier = generate_read_specifier(self.words_file_name)
            words_reader = SequentialInt32VectorReader(words_read_specifier)
        if self.likelihood_file_name is not None:
            likelihood_read_specifier = generate_read_specifier(self.likelihood_file_name)
            likelihood_reader = SequentialBaseFloatVectorReader(likelihood_read_specifier)
        try:
            while not reader.Done():
                utt = reader.Key()
                alignment = reader.Value()
                words = []
                if words_reader is not None:
                    words_utt = words_reader.Key()
                    if words_utt != utt:
                        raise Exception(
                            f"Mismatch in keys between {self.file_name} and {self.words_file_name}"
                        )
                    words = words_reader.Value()
                likelihoods = []
                if likelihood_reader is not None:
                    likelihood_utt = likelihood_reader.Key()
                    if likelihood_utt != utt:
                        raise Exception(
                            f"Mismatch in keys between {self.file_name} and {self.likelihood_file_name}"
                        )
                    likelihoods = likelihood_reader.Value()
                alignment = Alignment(
                    utt, alignment, words=words, per_frame_likelihoods=likelihoods
                )
                yield alignment
                reader.Next()
                if words_reader is not None:
                    words_reader.Next()
                if likelihood_reader is not None:
                    likelihood_reader.Next()
        finally:
            reader.Close()

    def __getitem__(self, item: str) -> Alignment:
        """Get alignment for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        words = []
        likelihoods = []
        if self.words_reader is not None:
            words = self.words_reader.Value(item)
        if self.likelihood_reader is not None:
            likelihoods = self.likelihood_reader.Value(item)
        return Alignment(
            item, self.random_reader.Value(item), words=words, per_frame_likelihoods=likelihoods
        )


class TranscriptionArchive:
    """
    Class for reading an archive or SCP of transcriptions

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(
        self,
        file_name: typing.Union[pathlib.Path, str],
        lm_scale: float = 1.0,
        acoustic_scale: float = 1.0,
    ):
        self.lattice_archive = LatticeArchive(file_name, determinized=True)
        self.lm_scale = lm_scale
        self.acoustic_scale = acoustic_scale

    def __del__(self):
        self.lattice_archive.close()

    def __iter__(self) -> typing.Generator[Alignment]:
        """Iterate over the utterance alignments in the archive"""
        reader = self.lattice_archive.sequential_reader
        try:
            while not reader.Done():
                utt = reader.Key()
                lattice = reader.Value()
                decoded = lattice_best_path(lattice, self.lm_scale, self.acoustic_scale)
                alignment, words, weight = GetLinearSymbolSequence(decoded)
                likelihood = -(weight.Value1() + weight.Value2()) / self.acoustic_scale
                transcription = Alignment(
                    utt, alignment, words, likelihood=likelihood, lattice=lattice
                )
                yield transcription
                reader.Next()
        finally:
            reader.Close()

    def __getitem__(self, item: str) -> Alignment:
        """Get alignment for a particular key from the archive file"""
        item = str(item)
        if not self.lattice_archive.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.lattice_archive.file_name}")

        lattice = self.lattice_archive.random_reader.Value(item)
        decoded = lattice_best_path(lattice, self.lm_scale, self.acoustic_scale)
        alignment, words, weight = GetLinearSymbolSequence(decoded)
        likelihood = -(weight.Value1() + weight.Value2()) / self.acoustic_scale
        transcription = Alignment(item, alignment, words, likelihood=likelihood, lattice=lattice)
        return transcription


class LatticeArchive:
    """
    Class for reading an archive or SCP of lattices

    Parameters
    ----------
    file_name: :class:`~pathlib.Path` or str
        Path to archive or SCP file to read from
    """

    def __init__(self, file_name: typing.Union[pathlib.Path, str], determinized: bool = True):
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.determinized = determinized
        self.read_specifier = generate_read_specifier(file_name)
        if self.determinized:
            self.random_reader = RandomAccessCompactLatticeReader(self.read_specifier)
        else:
            self.random_reader = RandomAccessLatticeReader(self.read_specifier)

    def close(self):
        if self.random_reader.IsOpen():
            self.random_reader.Close()

    @property
    def sequential_reader(
        self,
    ) -> typing.Union[SequentialCompactLatticeReader, SequentialLatticeReader]:
        """Sequential reader for lattices"""
        if self.determinized:
            return SequentialCompactLatticeReader(self.read_specifier)
        return SequentialLatticeReader(self.read_specifier)

    def __iter__(
        self,
    ) -> typing.Generator[typing.Tuple[str, typing.Union[Lattice, CompactLattice]]]:
        """Iterate over the utterance lattices in the archive"""
        reader = self.sequential_reader
        try:
            while not reader.Done():
                utt = reader.Key()
                lattice = reader.Value()
                yield utt, lattice
                reader.Next()
        finally:
            reader.Close()

    def __del__(self):
        self.close()

    def __getitem__(self, item: str) -> typing.Union[Lattice, CompactLattice]:
        """Get lattice for a particular key from the archive file"""
        item = str(item)
        if not self.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.file_name}")
        return self.random_reader.Value(item)



"""Module for evaluating alignments"""
from __future__ import annotations

import collections
import typing

from _kalpy.util import IntervalAlignment, align_intervals
from kalpy.gmm.data import CtmInterval, WordCtmInterval

if typing.TYPE_CHECKING:
    from kalpy.fstext.lexicon import LexiconCompiler


def compare_labels(
    reference_label: str,
    test_label: str,
    silence_phones: typing.Set[str],
    mapping: typing.Optional[typing.Dict[str, typing.Collection[str]]] = None,
) -> int:
    """
    Score two labels based on whether they match or count as the same phone based on the mapping

    Parameters
    ----------
    reference_label: str
        Reference label
    test_label: str
        Label to compare to reference label
    silence_phones: set[str]
        Label corresponding to optionally inserted silence
    mapping: Optional[dict[str, set[str]]]
        Mapping to form equivalence between different label sets

    Returns
    -------
    int
        0 if labels match or if they're in the mapping 2 otherwise. If one of the labels is the silence_phone, returns 10
    """
    if reference_label == test_label:
        return 0
    if mapping is not None and test_label in mapping:
        if isinstance(mapping[test_label], str):
            if mapping[test_label] == reference_label:
                return 0
        elif reference_label in mapping[test_label]:
            return 0
    ref = reference_label.lower()
    test = test_label.lower()
    if ref == test:
        return 0
    if reference_label in silence_phones or test_label in silence_phones:
        return 10
    return 2


def overlap_scoring(
    first_element: CtmInterval,
    second_element: CtmInterval,
    silence_phones: typing.Set[str],
    mapping: typing.Optional[typing.Dict[str, str]] = None,
) -> float:
    r"""
    Method to calculate overlap scoring

    .. math::

       Score = -(\lvert begin_{1} - begin_{2} \rvert + \lvert end_{1} - end_{2} \rvert + \begin{cases}
                0, & if label_{1} = label_{2} \\
                2, & otherwise
                \end{cases})

    See Also
    --------
    `Blog post <https://memcauliffe.com/update-on-montreal-forced-aligner-performance.html>`_
        For a detailed example that using this metric

    Parameters
    ----------
    first_element: :class:`~montreal_forced_aligner.data.CtmInterval`
        First CTM interval to compare
    second_element: :class:`~montreal_forced_aligner.data.CtmInterval`
        Second CTM interval
    silence_phones: set[str]
        Labels corresponding to optionally inserted silence
    mapping: Optional[dict[str, str]]
        Optional mapping of phones to treat as matches even if they have different symbols

    Returns
    -------
    float
        Score calculated as the negative sum of the absolute different in begin timestamps, absolute difference in end
        timestamps and the label score
    """
    begin_diff = abs(first_element.begin - second_element.begin)
    end_diff = abs(first_element.end - second_element.end)
    label_diff = compare_labels(first_element.label, second_element.label, silence_phones, mapping)
    return -1 * (begin_diff + end_diff + label_diff)


def fix_many_to_one_alignments(
    alignment: IntervalAlignment, custom_mapping: typing.Optional[typing.Dict[str, str]]
):
    test_keys = set(x for x in custom_mapping.keys() if " " in x)
    ref_keys = set()
    for val in custom_mapping.values():
        ref_keys.update(x for x in val if " " in x)
    new_ref = []
    new_test = []
    for i, (sa, sb) in enumerate(alignment.alignment):
        if i != 0:
            prev_sa, prev_sb = alignment.alignment[i - 1]
            ref_key = " ".join(x.label for x in [prev_sa, sa] if x.label != "-")
            test_key = " ".join(x.label for x in [prev_sb, sb] if x.label != "-")
            if (
                ref_key in ref_keys
                and test_key in custom_mapping
                and ref_key in custom_mapping[test_key]
            ):
                new_ref[-1].label = ref_key
                new_ref[-1].end = sa.end
                if sb.label != "-":
                    new_test.append(sb)
                continue
            if (
                test_key in test_keys
                and test_key in custom_mapping
                and ref_key in custom_mapping[test_key]
            ):
                new_test[-1].label = test_key
                new_test[-1].end = sb.end
                if sa.label != "-":
                    new_ref.append(sa)
                continue
        if sa.label != "-":
            new_ref.append(sa)
        if sb.label != "-":
            new_test.append(sb)
    return new_ref, new_test


def align_phones(
    ref: typing.List[CtmInterval],
    test: typing.List[CtmInterval],
    silence_phones: typing.Union[str, typing.Set[str]],
    ignored_phones: typing.Set[str] = None,
    custom_mapping: typing.Optional[typing.Dict[str, typing.Set[str]]] = None,
    debug: bool = False,
) -> typing.Tuple[
    float,
    float,
    typing.Dict[typing.Tuple[str, str], int],
    float,
    typing.List[typing.Dict[str, typing.Any]],
]:
    """
    Align phones based on how much they overlap and their phone label, with the ability to specify a custom mapping for
    different phone labels to be scored as if they're the same phone

    Parameters
    ----------
    ref: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of CTM intervals as reference
    test: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of CTM intervals to compare to reference
    silence_phones: str or set[str]
        Silence phones (these are ignored in the final calculation)
    ignored_phones: set[str], optional
        Phones that should be ignored in score calculations (silence phone is automatically added)
    custom_mapping: dict[str, set[str]], optional
        Mapping of phones to treat as matches even if they have different symbols
    debug: bool, optional
        Flag for logging extra information about alignments

    Returns
    -------
    float
        Score based on the average amount of overlap in phone intervals
    float
        Phone error rate
    dict[tuple[str, str], int]
        Dictionary of error pairs with their counts
    float
        Edit distance score of alignment
    list[dict[str, any]]
        Boundary errors
    """
    if isinstance(silence_phones, str):
        silence_phones = {silence_phones}
    if ignored_phones is None:
        ignored_phones = set()
    if not isinstance(ignored_phones, set):
        ignored_phones = set(ignored_phones)
    if custom_mapping is None:
        custom_mapping = {}
    else:
        for k, v in custom_mapping.items():
            if k in silence_phones:
                if isinstance(v, str):
                    silence_phones.add(v)
                else:
                    silence_phones.update(v)
            elif v & silence_phones:
                silence_phones.add(k)
    ignored_phones.update(silence_phones)
    try:
        alignment = align_intervals(ref, test, silence_phones, custom_mapping)
    except Exception:
        print(ref)
        print(test)
        raise
    if custom_mapping is not None:
        ref, test = fix_many_to_one_alignments(alignment, custom_mapping)
        alignment = align_intervals(ref, test, silence_phones, custom_mapping)
    overlap_count = 0
    overlap_sum = 0
    num_insertions = 0
    num_deletions = 0
    num_substitutions = 0
    errors = collections.Counter()
    boundary_errors = []
    for i, (sa, sb) in enumerate(alignment.alignment):
        if sa.label == "-":
            if sb.label not in ignored_phones:
                errors[(sa.label, sb.label)] += 1
                num_insertions += 1
            else:
                continue
        elif sb.label == "-":
            if sa.label not in ignored_phones:
                errors[(sa.label, sb.label)] += 1
                num_deletions += 1
            else:
                continue
        else:
            if i != 0:
                prev_sa, prev_sb = alignment.alignment[i - 1]
                if prev_sa.label != "-" and prev_sb.label != "-":
                    boundary_errors.append(
                        {
                            "following_reference_phone": sa.label,
                            "following_test_phone": sb.label,
                            "previous_reference_phone": prev_sa.label,
                            "previous_test_phone": prev_sb.label,
                            "boundary_error": round(sa.begin - sb.begin, 3),
                            "reference_boundary": round(sa.begin, 3),
                            "test_boundary": round(sb.begin, 3),
                        }
                    )
            if sa.label in ignored_phones:
                continue
            overlap_sum += (abs(sa.begin - sb.begin) + abs(sa.end - sb.end)) / 2
            overlap_count += 1
            if compare_labels(sa.label, sb.label, silence_phones, mapping=custom_mapping) > 0:
                num_substitutions += 1
                errors[(sa.label, sb.label)] += 1
    if overlap_count:
        score = round(overlap_sum / overlap_count, 3)
    else:
        score = None
    phone_error_rate = (num_insertions + num_deletions + (2 * num_substitutions)) / len(ref)
    if debug:
        import logging

        logger = logging.getLogger("mfa")
        if errors:
            logger.debug(
                f"PER: {phone_error_rate}\nErrors: {errors}\n{format_alignment(alignment)}"
            )
        else:
            logger.debug(f"PER: {phone_error_rate}\n{format_alignment(alignment)}")
    return score, phone_error_rate, errors, alignment.score, boundary_errors


def fix_unk_words(
    ref: typing.List[str],
    test: typing.List[CtmInterval],
    lexicon_compiler: LexiconCompiler,
) -> typing.List[WordCtmInterval]:
    """
    Takes in word-level alignments and looks up original label of unknown words

    Parameters
    ----------
    ref: list[str]
        Reference text
    test: list[:class:`~kalpy.gmm.data.WordCtmInterval`]
        Aligned word intervals with unknown tokens
    lexicon_compiler: LexiconCompiler
        Lexicon compiler to use for evaluating the identity of OOV items

    Returns
    -------
    list[:class:`~kalpy.gmm.data.WordCtmInterval`]
        Aligned words with unknown word tokens replaced with their original label
    """

    mapping = {}
    ref_intervals = []
    for w in ref:
        ref_intervals.append(WordCtmInterval(w, lexicon_compiler.to_int(w), []))
        if lexicon_compiler.to_int(w) == lexicon_compiler.to_int(lexicon_compiler.oov_word):
            if w not in mapping:
                mapping[w] = {lexicon_compiler.oov_word}

    alignment = align_intervals(ref_intervals, test, {lexicon_compiler.silence_word}, {})
    output_ctm = []
    for sa, sb in alignment.alignment:
        if sa.label == "-":
            output_ctm.append(sb)
        elif sb.label == "-":
            continue
        else:
            if sa.label != sb.label and sb.label == lexicon_compiler.oov_word:
                sb.label = sa.label
            output_ctm.append(sb)
    return output_ctm


def align_words(
    ref: typing.Union[typing.List[str], typing.List[CtmInterval]],
    test: typing.List[CtmInterval],
    silence_words: typing.Union[str, typing.Set[str]],
    ignored_words: typing.Set[str] = None,
    debug: bool = False,
) -> typing.Tuple[float, float, float]:
    """
    Align words based on how much their time points overlap and their label

    Parameters
    ----------
    ref: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of CTM intervals as reference
    test: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
        List of CTM intervals to compare to reference
    silence_words: str or set[str]
        Silence word (these are ignored in the final calculation)
    ignored_words: set[str], optional
        Words that should be ignored in score calculations (silence phone is automatically added)
    debug: bool, optional
        Flag for logging extra information about alignments

    Returns
    -------
    float
        Extra duration of new words
    float
        Word error rate
    float
        Aligned duration of found words
    """

    if isinstance(silence_words, str):
        silence_words = {silence_words}
    if ignored_words is None:
        ignored_words = set()
    if not isinstance(ignored_words, set):
        ignored_words = set(ignored_words)
    ignored_words.update(silence_words)
    ref_intervals = []
    for w in ref:
        if not isinstance(w, CtmInterval):
            ref_intervals.append(CtmInterval(0.0, 0.0, w))
        else:
            ref_intervals.append(w)
    alignment = align_intervals(ref_intervals, test, silence_words, {})

    num_insertions = 0
    num_deletions = 0
    num_substitutions = 0

    extra_duration = 0
    aligned_duration = 0
    for sa, sb in alignment.alignment:
        if sa.label == "-":
            if sb.label not in ignored_words:
                num_insertions += 1
                extra_duration += sb.end - sb.begin
            else:
                continue
        elif sb.label == "-":
            if sa not in ignored_words:
                num_deletions += 1
            else:
                continue
        else:
            if sa.label in ignored_words:
                continue
            if sa.label != sb.label:
                num_substitutions += 1
            else:
                aligned_duration += sb.end - sb.begin
    word_error_rate = (num_insertions + num_deletions + (2 * num_substitutions)) / len(ref)
    if debug:
        import logging

        logger = logging.getLogger("mfa")
        logger.debug(
            f"{format_alignment(alignment)}\nExtra word duration: {extra_duration}\nWER: {word_error_rate}"
        )
    return extra_duration, word_error_rate, aligned_duration


def format_alignment(alignment: IntervalAlignment) -> str:
    """Format the alignment prettily into a string.

    Adapted largely from
    `Bio python's pairwise2 format_alignment <https://github.com/biopython/biopython/blob/master/Bio/pairwise2.py#L1348>`_

    Parameters
    ----------
    alignment: IntervalAlignment
        Alignment to format into human-readable form

    Returns
    -------
    str
        Formatted alignment string
    """
    begin = 0
    end = len(alignment)
    align1 = alignment.reference
    align2 = alignment.test
    full_sequences = False
    align_begin = begin
    align_end = end
    start1 = start2 = ""
    start_m = begin  # Begin of match line (how many spaces to include)
    # For local alignments:
    if not full_sequences and (begin != 0 or end != len(align1)):
        # Calculate the actual start positions in the un-aligned sequences
        # This will only work if the gap symbol is '-' or ['-']!
        start1 = str(len(align1[:begin]) - align1[:begin].count("-") + 1) + " "
        start2 = str(len(align2[:begin]) - align2[:begin].count("-") + 1) + " "
        start_m = max(len(start1), len(start2))
    elif full_sequences:
        start_m = 0
        begin = 0
        end = len(align1)

    s1_line = ["{:>{width}}".format(start1, width=start_m)]  # seq1 line
    m_line = [" " * start_m]  # match line
    s2_line = ["{:>{width}}".format(start2, width=start_m)]  # seq2 line

    for n, (r, t) in enumerate(alignment.alignment[begin:end]):
        # Since list elements can be of different length, we center them,
        # using the maximum length of the two compared elements as width
        a = r.label + " "
        b = t.label + " "
        m_len = max(len(a), len(b))
        s1_line.append("{:^{width}}".format(a, width=m_len))
        s2_line.append("{:^{width}}".format(b, width=m_len))
        if full_sequences and (n < align_begin or n >= align_end):
            m_line.append("{:^{width}}".format(" ", width=m_len))  # space
            continue
        if a == b:
            m_line.append("{:^{width}}".format("|", width=m_len))  # match
        elif a.strip() == "-" or b.strip() == "-":
            m_line.append("{:^{width}}".format(" ", width=m_len))  # gap
        else:
            m_line.append("{:^{width}}".format(".", width=m_len))  # mismatch

    s2_line.append(f"\n  Score={alignment.score:g}\n")
    return "\n".join(["".join(s1_line), "".join(m_line), "".join(s2_line)])



from _kalpy.tree import ContextDependency
from _kalpy.util import ReadKaldiObject


def test_transition_model(mono_tree_path):
    tree = ContextDependency()
    ReadKaldiObject(str(mono_tree_path), tree)
    assert tree.NumPdfs() > 0



"""Classes for computing MFCC features"""
from __future__ import annotations

import logging
import pathlib
import typing

import librosa
import numpy as np

from _kalpy import feat
from _kalpy.matrix import CompressedMatrix, FloatMatrix, FloatVector
from _kalpy.util import BaseFloatMatrixWriter, CompressedMatrixWriter
from kalpy.data import Segment
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.mfcc")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class MfccComputer:
    """
    Class for computing MFCC features

    Parameters
    ----------
    sample_frequency: float
        Sample rate to use in generating features, audio will be resampled as necessary,
        defaults to 16000
    frame_length: int
        Frame length in milliseconds, defaults to 25
    frame_shift: int
        Frame shift in milliseconds, defaults to 10
    dither: float
        Dithering to use for generating deterministic features while avoiding numerical zeros,
        defaults to -1
    preemphasis_coefficient: float
        Pre-emphasis coefficient to use prior to feature calculation, defaults to 0.97
    remove_dc_offset: bool
        Flag for removing DC offset, defaults to True
    window_type: str
        Type of window to use in generating frames, defaults to "povey"
    round_to_power_of_two: bool
        Flag for using a window based on the power of two in FFT calculation for efficiency,
        defaults to True
    blackman_coeff: float
        Coefficient to use when `window_type` is "blackman", defaults to 0.42
    snip_edges: bool
        Flag for whether edges of segments should be cutoff to ensure no zero padding in frames,
        defaults to True
    max_feature_vectors: int
        Maximum number of vectors to store in memory for VTLN calculation, defaults to -1
    num_mel_bins: int
        Number of mel frequency bins to use in calculating MFCCs, defaults to 23
    low_frequency: float
        Lowest frequency for the mel spectrum, defaults to 20
    high_frequency: float
        Highest frequency for the mel spectrum, defaults to 7800
    vtln_low: float
        VTLN lower cutoff of warping function, defaults to 100
    vtln_high: float
        VTLN upper cutoff of warping function if negative,
        added to the Nyquist frequency to get the cutoff, defaults to -500
    num_coefficients: int
        Number of MFCC coefficients, defaults to 13
    use_energy: bool
        Use energy of frame in place of the zeroth MFCC coefficient, defaults to True
    energy_floor: float
        Energy floor for MFCC, defaults to 0, set to 1.0 or 0.1 if dithering is disabled
    raw_energy: bool
        Flag for computing energy before pre-emphasis and windowing, defaults to True
    cepstral_lifter: float
        Scaling factor on cepstra for HTK compatibility, defaults to 22.0
    htk_compatibility: bool
        Flag for generating features in HTK format
    """

    def __init__(
        self,
        sample_frequency: float = 16000,
        frame_length: int = 25,
        frame_shift: int = 10,
        dither: float = 0.0,
        preemphasis_coefficient: float = 0.97,
        remove_dc_offset: bool = True,
        window_type: str = "povey",
        round_to_power_of_two: bool = True,
        blackman_coeff: float = 0.42,
        snip_edges: bool = False,
        max_feature_vectors: int = -1,
        num_mel_bins: int = 23,
        low_frequency: float = 20,
        high_frequency: float = 7800,
        vtln_low: float = 100,
        vtln_high: float = -500,
        num_coefficients: int = 13,
        use_energy: bool = True,
        energy_floor: float = 1.0,
        raw_energy: bool = True,
        cepstral_lifter: float = 22.0,
        htk_compatibility: bool = False,
        allow_downsample: bool = True,
        allow_upsample: bool = True,
    ):
        self.frame_length = frame_length
        self._frame_shift = frame_shift
        self.dither = dither
        self.preemphasis_coefficient = preemphasis_coefficient
        self.sample_frequency = sample_frequency
        self.remove_dc_offset = remove_dc_offset
        self.window_type = window_type
        self.round_to_power_of_two = round_to_power_of_two
        self.blackman_coeff = blackman_coeff
        self.snip_edges = snip_edges
        self.max_feature_vectors = max_feature_vectors
        self.allow_downsample = allow_downsample
        self.allow_upsample = allow_upsample
        self.low_frequency = low_frequency
        self.high_frequency = high_frequency
        self.vtln_low = vtln_low
        self.vtln_high = vtln_high
        self.num_mel_bins = num_mel_bins
        self.cepstral_lifter = cepstral_lifter
        self.num_coefficients = num_coefficients
        self.energy_floor = energy_floor
        self.use_energy = use_energy
        self.raw_energy = raw_energy
        self.htk_compatibility = htk_compatibility

    @property
    def parameters(self):
        return {
            "sample_frequency": self.sample_frequency,
            "frame_length": self.frame_length,
            "frame_shift": self._frame_shift,
            "dither": self.dither,
            "preemphasis_coefficient": self.preemphasis_coefficient,
            "remove_dc_offset": self.remove_dc_offset,
            "window_type": self.window_type,
            "round_to_power_of_two": self.round_to_power_of_two,
            "blackman_coeff": self.blackman_coeff,
            "snip_edges": self.snip_edges,
            "max_feature_vectors": self.max_feature_vectors,
            "num_mel_bins": self.num_mel_bins,
            "low_frequency": self.low_frequency,
            "high_frequency": self.high_frequency,
            "vtln_low": self.vtln_low,
            "vtln_high": self.vtln_high,
            "num_coefficients": self.num_coefficients,
            "use_energy": self.use_energy,
            "energy_floor": self.energy_floor,
            "raw_energy": self.raw_energy,
            "cepstral_lifter": self.cepstral_lifter,
            "htk_compatibility": self.htk_compatibility,
            "allow_downsample": self.allow_downsample,
            "allow_upsample": self.allow_upsample,
        }

    @property
    def frame_shift(self):
        return round(self._frame_shift / 1000, 3)

    @property
    def mfcc_obj(self):

        frame_opts = feat.FrameExtractionOptions()
        frame_opts.frame_length_ms = self.frame_length
        frame_opts.frame_shift_ms = self._frame_shift
        frame_opts.dither = self.dither
        frame_opts.preemph_coeff = self.preemphasis_coefficient
        frame_opts.samp_freq = self.sample_frequency
        frame_opts.remove_dc_offset = self.remove_dc_offset
        frame_opts.window_type = self.window_type
        frame_opts.round_to_power_of_two = self.round_to_power_of_two
        frame_opts.blackman_coeff = self.blackman_coeff
        frame_opts.snip_edges = self.snip_edges
        frame_opts.max_feature_vectors = self.max_feature_vectors
        frame_opts.allow_downsample = self.allow_downsample
        frame_opts.allow_upsample = self.allow_upsample

        mel_opts = feat.MelBanksOptions(self.num_mel_bins)
        mel_opts.low_freq = self.low_frequency
        mel_opts.high_freq = self.high_frequency
        mel_opts.vtln_low = self.vtln_low
        mel_opts.vtln_high = self.vtln_high

        mfcc_opts = feat.MfccOptions()
        mfcc_opts.frame_opts = frame_opts
        mfcc_opts.mel_opts = mel_opts
        mfcc_opts.cepstral_lifter = self.cepstral_lifter
        mfcc_opts.num_ceps = self.num_coefficients
        mfcc_opts.use_energy = self.use_energy
        mfcc_opts.energy_floor = self.energy_floor
        mfcc_opts.raw_energy = self.raw_energy
        mfcc_opts.htk_compat = self.htk_compatibility
        return feat.Mfcc(mfcc_opts)

    def compute_mfccs(
        self,
        segment: typing.Union[Segment, np.ndarray],
    ) -> np.ndarray:
        """
        Compute MFCC features for a segment

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate MFCCs

        Returns
        -------
        :class:`numpy.ndarray`
            Feature matrix for the segment
        """
        mfccs = self.compute_mfccs_for_export(segment, compress=False)
        return mfccs.numpy()

    def compute_mfccs_for_export(
        self, segment: typing.Union[Segment, np.ndarray, FloatVector], compress: bool = True
    ) -> FloatMatrix:
        """
        Generate MFCCs for exporting to a kaldi archive

        Parameters
        ----------
        segment: :class:`~kalpy.feat.mfcc.Segment`
            Acoustic segment to generate MFCCs
        compress: bool, defaults to True
            Flag for whether returned matrix should be compressed

        Returns
        -------
        :class:`_kalpy.matrix.FloatMatrix`
            Feature matrix for the segment
        """
        if isinstance(segment, Segment):
            wave = segment.kaldi_wave
        else:
            wave = segment
            if isinstance(wave, np.ndarray) and np.max(wave) < 1.0:
                wave = wave * 32768

        mfccs = self.mfcc_obj.compute(wave)
        if compress:
            mfccs = CompressedMatrix(mfccs)
        return mfccs

    def export_feats(
        self,
        file_name: typing.Union[pathlib.Path, str],
        segments: typing.Iterable[typing.Tuple[str, Segment]],
        write_scp: bool = False,
        compress: bool = True,
    ) -> None:
        """
        Export features to a kaldi archive file (i.e., mfccs.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        segments: dict[str, :class:`kalpy.feat.mfcc.Segment`]
            Mapping of utterance IDs to Segment objects
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        compress: bool
            Flag for whether to export features as a compressed archive
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        logger.debug(f"Writing to: {write_specifier}")
        if compress:
            writer = CompressedMatrixWriter(write_specifier)
        else:
            writer = BaseFloatMatrixWriter(write_specifier)
        for key, segment in segments:
            feats = self.compute_mfccs_for_export(segment, compress)
            writer.Write(str(key), feats)
        writer.Close()



"""Classes for computing MFCC features"""
from __future__ import annotations

import logging
import pathlib
import typing

from _kalpy.ivector import ComputeVadEnergy, VadEnergyOptions
from _kalpy.matrix import FloatMatrix, FloatVector
from _kalpy.util import BaseFloatVectorWriter
from kalpy.feat.data import FeatureArchive
from kalpy.utils import generate_write_specifier

logger = logging.getLogger("kalpy.vad")
logger.setLevel(logging.DEBUG)
logger.write = lambda msg: logger.info(msg) if msg != "\n" else None
logger.flush = lambda: None


class VadComputer:
    """
    Class for computing VAD features

    Parameters
    ----------
    energy_threshold: float
        Constant term in energy threshold for MFCC0 for VAD
    energy_mean_scale: float
        If this is set to s, to get the actual threshold we let m be the mean log-energy of the file,
        and use s*m + energy_threshold
    frames_context: int
        Number of frames of context on each side of central frame, in window for which energy is monitored
    proportion_threshold: float
        Parameter controlling the proportion of frames within the window that need to have more energy than the threshold
    """

    def __init__(
        self,
        energy_threshold: float = 5.0,
        energy_mean_scale: float = 0.5,
        frames_context: int = 0,
        proportion_threshold: float = 0.6,
    ):
        self.energy_threshold = energy_threshold
        self.energy_mean_scale = energy_mean_scale
        self.frames_context = frames_context
        self.proportion_threshold = proportion_threshold
        self.num_done = 0
        self.num_skipped = 0
        self.options = VadEnergyOptions()
        self.options.vad_energy_threshold = self.energy_threshold
        self.options.vad_energy_mean_scale = self.energy_mean_scale
        self.options.vad_frames_context = self.frames_context
        self.options.vad_proportion_threshold = self.proportion_threshold

    def compute_vad(
        self,
        features: FloatMatrix,
    ) -> FloatVector:
        """
        Compute VAD features for a segment

        Parameters
        ----------
        features: :class:`~_kalpy.matrix.FloatMatrix`
            Feature matrix to compute VAD

        Returns
        -------
        :class:`~_kalpy.matrix.FloatVector`
            VAD for each
        """
        return ComputeVadEnergy(self.options, features)

    def compute_vads(
        self,
        feature_archive: FeatureArchive,
    ) -> typing.Tuple[str, FloatVector]:
        """
        Compute VAD features for a segment

        Parameters
        ----------
        feature_archive: :class:`~kalpy.feat.data.FeatureArchive`
            Archive of features to compute VAD

        Yields
        -------
        str, :class:`~_kalpy.matrix.FloatVector`
            VAD for each
        """
        for utterance_id, feats in feature_archive:
            if feats.NumRows() == 0:
                logger.debug(f"Skipping {utterance_id} due to empty features.")
                self.num_skipped += 1
                continue
            vad = self.compute_vad(feats)
            self.num_done += 1
            yield utterance_id, vad
        logger.info(f"Done {self.num_done} utterances, skipped {self.num_skipped}.")

    def export_vad(
        self,
        file_name: typing.Union[pathlib.Path, str],
        feature_archive: FeatureArchive,
        write_scp: bool = False,
        callback: typing.Callable = None,
    ):
        """
        Export features to a kaldi archive file (i.e., mfccs.ark)

        Parameters
        ----------
        file_name: :class:`~pathlib.Path` or str
            Archive file path to export to
        feature_archive: :class:`~kalpy.feat.data.FeatureArchive`
            Archive of features to compute VAD
        write_scp: bool
            Flag for whether an SCP file should be generated as well
        callback: typing.Callable
            Flag for yielding processed utterance IDs
        """
        write_specifier = generate_write_specifier(file_name, write_scp)
        logger.debug(f"Writing to: {write_specifier}")
        writer = BaseFloatVectorWriter(write_specifier)
        for key, vad in self.compute_vads(feature_archive):
            if callback:
                callback(key)
            writer.Write(str(key), vad)
        writer.Close()



"""Classes for higher level alignment"""
from __future__ import annotations

import typing

from _kalpy.gmm import gmm_interpolate_boundary_fast
from _kalpy.hmm import SplitToPhones
from _kalpy.matrix import DoubleMatrix, FloatMatrix, FloatSubMatrix
from kalpy.data import Segment
from kalpy.decoder.training_graphs import TrainingGraphCompiler
from kalpy.exceptions import AlignerError
from kalpy.gmm.align import GmmAligner
from kalpy.gmm.data import CtmInterval
from kalpy.models import AcousticModel

if typing.TYPE_CHECKING:
    from kalpy.fstext.lexicon import LexiconCompiler
    from kalpy.gmm.data import Alignment, HierarchicalCtm
    from kalpy.utterance import Utterance


class KalpyAligner:
    """
    Aligner class for generating alignments from a :class:`~kalpy.models.AcousticModel` and
    :class:`~kalpy.fstext.lexicon.LexiconCompiler`.

    Parameters
    ----------
    acoustic_model: :class:`~kalpy.models.AcousticModel`
        Acoustic model
    lexicon_compiler: :class:`~kalpy.fstext.lexicon.LexiconCompiler` or dict[Any, :class:`~kalpy.fstext.lexicon.LexiconCompiler`]
        Single lexicon compiler to use for all utterances or a dictionary of different lexicon compilers to use
        depending on the utterance
    beam: int
        Size of the beam to use in decoding, defaults to 10
    retry_beam : int
        Size of the beam to use in decoding if it fails with the initial beam width, defaults to 40
    transition_scale : float
        Transition scale, defaults to 1.0
    acoustic_scale : float
        Acoustic scale, defaults to 0.1
    self_loop_scale : float
        Self-loop scale, defaults to 0.1
    boost_silence : float
        Factor to boost silence probabilities, 1.0 is no boost or reduction
    careful : bool
        Flag for extra error checking on reaching final state, defaults to False
    """

    def __init__(
        self,
        acoustic_model: AcousticModel,
        lexicon_compiler: typing.Union[LexiconCompiler, typing.Dict[typing.Any, LexiconCompiler]],
        beam: int = 10,
        retry_beam: int = 40,
        transition_scale: float = 1.0,
        acoustic_scale: float = 0.1,
        self_loop_scale: float = 0.1,
        boost_silence: float = 1.0,
        careful: bool = False,
    ):
        self.acoustic_model = acoustic_model
        if isinstance(lexicon_compiler, dict) and len(lexicon_compiler) == 1:
            lexicon_compiler = next(iter(lexicon_compiler.values()))
        self.lexicon_compiler = lexicon_compiler
        self.has_multiple_lexicons = isinstance(self.lexicon_compiler, dict)

        if self.has_multiple_lexicons:
            self.graph_compiler = {}
            for k, v in self.lexicon_compiler.items():
                self.graph_compiler[k] = TrainingGraphCompiler(
                    self.acoustic_model.model_path,
                    self.acoustic_model.tree_path,
                    v,
                )
        else:
            self.graph_compiler = TrainingGraphCompiler(
                self.acoustic_model.model_path,
                self.acoustic_model.tree_path,
                self.lexicon_compiler,
            )
        self.beam = beam
        self.retry_beam = retry_beam
        self.transition_scale = transition_scale
        self.acoustic_scale = acoustic_scale
        self.self_loop_scale = self_loop_scale
        self.boost_silence = boost_silence
        self.careful = careful
        self.aligner = GmmAligner(
            self.acoustic_model.model_path,
            beam=beam,
            retry_beam=retry_beam,
            transition_scale=transition_scale,
            acoustic_scale=acoustic_scale,
            self_loop_scale=self_loop_scale,
            careful=careful,
        )
        self.ali_aligner = None
        if self.acoustic_model.alignment_model_path != self.acoustic_model.model_path:
            self.ali_aligner = GmmAligner(
                self.acoustic_model.alignment_model_path,
                beam=beam,
                retry_beam=retry_beam,
                transition_scale=transition_scale,
                acoustic_scale=acoustic_scale,
                self_loop_scale=self_loop_scale,
                careful=careful,
            )
        if self.boost_silence != 1.0:
            if self.has_multiple_lexicons:
                silence_symbols = next(iter(self.lexicon_compiler.values())).silence_symbols
            else:
                silence_symbols = self.lexicon_compiler.silence_symbols
            if self.ali_aligner is not None:
                self.ali_aligner.boost_silence(boost_silence, silence_symbols)
            self.aligner.boost_silence(boost_silence, silence_symbols)

    def _align_utterance(
        self,
        utterance: Utterance,
        cmvn: DoubleMatrix = None,
        fmllr_trans: FloatMatrix = None,
        dictionary_id: typing.Any = None,
    ) -> Alignment:
        """
        Internal function for generating an :class:`~kalpy.gmm.data.Alignment` from an :class:`~kalpy.utterance.Utterance`

        Parameters
        ----------
        utterance: :class:`~kalpy.utterance.Utterance`
            Utterance to align
        cmvn: :class:`~_kalpy.matrix.DoubleMatrix`, optional
            CMVN transformation to use, if not provided, a CMVN transform will be applied based on the utterance features
        fmllr_trans: :class:`~_kalpy.matrix.FloatMatrix`, optional
            Feature transformation matrix for a speaker
        dictionary_id: int or str, optional
            Identifier for which lexicon compiler to use, if not specified, defaults to using the first lexicon compiler

        Raises
        ------
        :class:`~kalpy.exceptions.AlignerError`
            Raised when no alignment is generated

        Returns
        -------
        :class:`~kalpy.gmm.data.Alignment`
            Alignment object with list of transition IDs and other information for the utterance
        """
        feats = utterance.generate_features(
            self.acoustic_model, fmllr_trans=fmllr_trans, cmvn=cmvn
        )
        graph_compiler = self.graph_compiler
        if self.has_multiple_lexicons:
            if dictionary_id is None:
                dictionary_id = next(iter(self.lexicon_compiler.keys()))
            graph_compiler = self.graph_compiler[dictionary_id]

        fst = graph_compiler.compile_fst(utterance.transcript)
        aligner = (
            self.ali_aligner
            if fmllr_trans is None and self.ali_aligner is not None
            else self.aligner
        )
        alignment = aligner.align_utterance(fst, feats)
        if alignment is None:
            raise AlignerError(
                f"Could not align the file with the current beam size ({aligner.beam}, "
                "please try increasing the beam size via `--beam X`"
            )
        return alignment

    def align_utterance(
        self,
        utterance: Utterance,
        cmvn: DoubleMatrix = None,
        fmllr_trans: FloatMatrix = None,
        dictionary_id: typing.Any = None,
    ) -> HierarchicalCtm:
        """
        Function for generating an :class:`~kalpy.gmm.data.Alignment` from an :class:`~kalpy.utterance.Utterance`

        Parameters
        ----------
        utterance: :class:`~kalpy.utterance.Utterance`
            Utterance to align
        cmvn: :class:`~_kalpy.matrix.DoubleMatrix`, optional
            CMVN transformation to use, if not provided, a CMVN transform will be applied based on the utterance features
        fmllr_trans: :class:`~_kalpy.matrix.FloatMatrix`, optional
            Feature transformation matrix for a speaker
        dictionary_id: int or str, optional
            Identifier for which lexicon compiler to use, if not specified, defaults to using the first lexicon compiler

        Raises
        ------
        :class:`~kalpy.exceptions.AlignerError`
            Raised when no alignment is generated

        Returns
        -------
        :class:`~kalpy.gmm.data.HierarchicalCtm`
            Hierarchical CTM object with word and phone intervals for the utterance
        """
        lexicon_compiler = self._get_lexicon_compiler(dictionary_id)
        aligner = (
            self.ali_aligner
            if fmllr_trans is None and self.ali_aligner is not None
            else self.aligner
        )
        alignment = self._align_utterance(utterance, cmvn, fmllr_trans, dictionary_id)
        phone_intervals = alignment.generate_ctm(
            aligner.transition_model,
            lexicon_compiler.phone_table,
            self.acoustic_model.mfcc_computer.frame_shift,
        )
        ctm = lexicon_compiler.phones_to_pronunciations(
            alignment.words, phone_intervals, transcription=False, text=utterance.transcript
        )
        ctm.likelihood = alignment.likelihood
        ctm.update_utterance_boundaries(utterance.segment.begin, utterance.segment.end)
        return ctm

    def _get_lexicon_compiler(self, dictionary_id: typing.Any = None) -> LexiconCompiler:
        """
        Internal function for looking up the lexicon compiler for the specified key

        Parameters
        ----------
        dictionary_id: int or str, optional
            Key to look up lexicon compiler

        Returns
        -------
        :class:`~kalpy.fstext.lexicon.LexiconCompiler`
            Lexicon compiler
        """
        lexicon_compiler = self.lexicon_compiler
        if self.has_multiple_lexicons:
            if dictionary_id is None:
                dictionary_id = next(iter(self.lexicon_compiler.keys()))
            lexicon_compiler = self.lexicon_compiler[dictionary_id]
        return lexicon_compiler

    def fine_tune_alignments(
        self,
        utterance: Utterance,
        alignment: Alignment = None,
        boundary_tolerance: typing.Optional[float] = None,
        cmvn: DoubleMatrix = None,
        fmllr_trans: FloatMatrix = None,
        dictionary_id: typing.Any = None,
    ) -> HierarchicalCtm:
        """
        Function for finetuning an :class:`~kalpy.gmm.data.Alignment` for an :class:`~kalpy.utterance.Utterance`

        Parameters
        ----------
        utterance: :class:`~kalpy.utterance.Utterance`
            Utterance to align
        alignment: :class:`~kalpy.gmm.data.Alignment`, optional
            Alignment to finetune, if not specified, alignment will be generated
        boundary_tolerance: float, optional
            The range around a boundary that it is allowed to be moved in seconds,
            if not specified, it will default to two frames from the acoustic model features, one frame on either side
        cmvn: :class:`~_kalpy.matrix.DoubleMatrix`, optional
            CMVN transformation to use, if not provided, a CMVN transform will be applied based on the utterance features
        fmllr_trans: :class:`~_kalpy.matrix.FloatMatrix`, optional
            Feature transformation matrix for a speaker
        dictionary_id: int or str, optional
            Identifier for which lexicon compiler to use, if not specified, defaults to using the first lexicon compiler

        Raises
        ------
        :class:`~kalpy.exceptions.AlignerError`
            Raised when no alignment is generated

        Returns
        -------
        :class:`~kalpy.gmm.data.HierarchicalCtm`
            Hierarchical CTM object with word and phone intervals for the utterance
        """
        lexicon_compiler = self._get_lexicon_compiler(dictionary_id)
        aligner = (
            self.ali_aligner
            if fmllr_trans is None and self.ali_aligner is not None
            else self.aligner
        )
        if alignment is None:
            alignment = self._align_utterance(
                utterance, cmvn=cmvn, fmllr_trans=fmllr_trans, dictionary_id=dictionary_id
            )
        split = SplitToPhones(aligner.transition_model, alignment.alignment)
        phone_intervals = []
        phone_start = 0.0
        original_start = 0.0
        feature_padding = 0.04
        if boundary_tolerance is None:
            boundary_tolerance = self.acoustic_model.frame_shift_seconds * 2
        for i, s in enumerate(split):
            phone_id = aligner.transition_model.TransitionIdToPhone(s[0])
            duration = len(s) * self.acoustic_model.frame_shift_seconds
            original_end = original_start + duration
            boundary = original_end
            label = lexicon_compiler.phone_table.find(phone_id)
            if i != len(split) - 1:
                feature_segment_begin = max(
                    round(boundary - feature_padding, 4),
                    0,
                )
                feature_segment_end = min(
                    round(boundary + feature_padding, 4),
                    utterance.segment.end,
                )
                following_phone_duration = (
                    len(split[i + 1]) * self.acoustic_model.frame_shift_seconds
                )
                previous_phone_offset_window = min(boundary_tolerance, duration / 2) / 2
                following_phone_offset_window = (
                    min(boundary_tolerance, following_phone_duration / 2) / 2
                )
                begin_offset = round(
                    max(boundary - previous_phone_offset_window - feature_segment_begin, 0.0), 4
                )
                end_offset = round(
                    min(
                        boundary + following_phone_offset_window - feature_segment_begin,
                        feature_segment_end - feature_segment_begin,
                    ),
                    4,
                )
                seg = Segment(
                    utterance.segment.file_path, feature_segment_begin, feature_segment_end
                )
                feats = self.acoustic_model.generate_features_for_fine_tune(
                    seg, cmvn=cmvn, fmllr_trans=fmllr_trans
                )
                begin_index = int(round(begin_offset * 1000))
                end_index = int(round(end_offset * 1000))
                sub_matrix = FloatSubMatrix(
                    feats, begin_index, end_index - begin_index, 0, feats.NumCols()
                )
                feats = FloatMatrix(sub_matrix)
                previous_transition_id = s[-1]
                following_transition_id = split[i + 1][0]
                new_boundary_index = gmm_interpolate_boundary_fast(
                    aligner.acoustic_model,
                    self.acoustic_model.transition_model,
                    feats,
                    previous_transition_id,
                    following_transition_id,
                )
                boundary = round(
                    feature_segment_begin + begin_offset + (new_boundary_index * 0.001), 3
                )
            confidence = 0.0
            phone_intervals.append(
                CtmInterval(
                    round(phone_start, 3),
                    round(boundary, 3),
                    label,
                    phone_id,
                    confidence,
                )
            )
            phone_start = boundary
            original_start = original_end
        ctm = lexicon_compiler.phones_to_pronunciations(
            alignment.words, phone_intervals, transcription=False, text=utterance.transcript
        )
        ctm.likelihood = alignment.likelihood
        ctm.update_utterance_boundaries(utterance.segment.begin, utterance.segment.end)
        return ctm



import pytest

from kalpy.data import KaldiMapping
from kalpy.decoder.decode_graph import DecodeGraphCompiler
from kalpy.feat.data import FeatureArchive
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.gmm.data import AlignmentArchive, LatticeArchive
from kalpy.gmm.decode import GmmDecoder
from kalpy.lm.rescore import LmRescorer


@pytest.mark.order(3)
def test_decode(mono_tree_path, mono_model_path, dictionary_path, mono_temp_dir, lm_path):
    lc = LexiconCompiler(position_dependent_phones=False, disambiguation=True)
    lc.load_pronunciations(dictionary_path)

    gc = DecodeGraphCompiler(mono_model_path, mono_tree_path, lc, arpa_path=lm_path)
    cmvn_file_name = mono_temp_dir.joinpath("cmvn.ark")
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        mono_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        deltas=True,
    )
    aligner = GmmDecoder(mono_model_path, gc.hclg_fst, beam=1000)
    for alignment in aligner.decode_utterances(feature_archive):
        assert alignment.utterance_id == "1-1"
        assert len(alignment.alignment) == 2672
        assert alignment.per_frame_likelihoods.numpy().shape[0] == 2672
        ctm = alignment.generate_ctm(aligner.transition_model, lc.phone_table)
        assert len(ctm) > 0


@pytest.mark.order(3)
def test_decode_sat_first_pass(
    sat_tree_path,
    sat_align_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    lm_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones, disambiguation=True)
    lc.load_pronunciations(sat_dictionary_path)
    hclg_path = sat_temp_dir.joinpath("hclg.fst")

    gc = DecodeGraphCompiler(sat_align_model_path, sat_tree_path, lc)
    gc.load_from_file(hclg_path)
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    lattice_file_name = sat_temp_dir.joinpath("lat.ark")
    alignment_file_name = sat_temp_dir.joinpath("ali_decode.ark")
    word_file_name = sat_temp_dir.joinpath("words.ark")
    utt2spk = KaldiMapping()
    textgrid_name = sat_temp_dir.joinpath("first_pass_decode.TextGrid")
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        splices=True,
    )
    decoder = GmmDecoder(sat_align_model_path, gc.hclg_fst)
    decoder.boost_silence(20.0, lc.silence_symbols)
    decoder.export_lattices(
        lattice_file_name,
        feature_archive,
        word_file_name=word_file_name,
        alignment_file_name=alignment_file_name,
    )
    assert lattice_file_name.exists()
    assert alignment_file_name.exists()
    alignment_archive = AlignmentArchive(alignment_file_name, words_file_name=word_file_name)
    alignment = alignment_archive["1-1"]
    assert len(alignment.alignment) == 2672
    intervals = alignment.generate_ctm(decoder.transition_model, lc.phone_table)
    ctm = lc.phones_to_pronunciations(alignment.words, intervals)
    ctm.export_textgrid(textgrid_name, file_duration=26.72)


@pytest.mark.order(5)
def test_decode_sat_second_pass(
    sat_tree_path,
    sat_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    lm_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones, disambiguation=True)
    lc.load_pronunciations(sat_dictionary_path)

    hclg_path = sat_temp_dir.joinpath("hclg.fst")
    gc = DecodeGraphCompiler(sat_model_path, sat_tree_path, lc)
    gc.load_from_file(hclg_path)
    cmvn_file_name = sat_temp_dir.joinpath("cmvn.ark")
    trans_file_name = sat_temp_dir.joinpath("trans_decode.ark")
    word_file_name = sat_temp_dir.joinpath("words.ark")
    textgrid_name = sat_temp_dir.joinpath("second_pass_decode.TextGrid")
    lattice_file_name = sat_temp_dir.joinpath("lat_second_pass.ark")
    alignment_file_name = sat_temp_dir.joinpath("ali_decode_second_pass.ark")
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    feature_archive = FeatureArchive(
        sat_temp_dir.joinpath("mfccs.ark"),
        utt2spk=utt2spk,
        cmvn_file_name=cmvn_file_name,
        lda_mat_file_name=sat_lda_mat_path,
        transform_file_name=trans_file_name,
        splices=True,
    )
    decoder = GmmDecoder(sat_model_path, gc.hclg_fst)
    decoder.boost_silence(20.0, lc.silence_symbols)
    decoder.export_lattices(
        lattice_file_name,
        feature_archive,
        word_file_name=word_file_name,
        alignment_file_name=alignment_file_name,
    )
    assert lattice_file_name.exists()
    assert alignment_file_name.exists()
    alignment_archive = AlignmentArchive(alignment_file_name, words_file_name=word_file_name)
    alignment = alignment_archive["1-1"]
    assert len(alignment.alignment) == 2672
    intervals = alignment.generate_ctm(decoder.transition_model, lc.phone_table)
    ctm = lc.phones_to_pronunciations(alignment.words, intervals)
    ctm.export_textgrid(textgrid_name, file_duration=26.72)


@pytest.mark.order(6)
def test_decode_sat_lm_rescore(
    sat_tree_path,
    sat_model_path,
    sat_lda_mat_path,
    sat_dictionary_path,
    sat_temp_dir,
    sat_phones,
    lm_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones, disambiguation=True)
    lc.load_pronunciations(sat_dictionary_path)

    hclg_path = sat_temp_dir.joinpath("hclg.fst")
    g_path = sat_temp_dir.joinpath("g.fst")
    gc = DecodeGraphCompiler(sat_model_path, sat_tree_path, lc)
    gc.load_from_file(hclg_path)
    gc.load_g_from_file(g_path)
    gc.compile_g_carpa(lm_path, sat_temp_dir.joinpath("g.carpa"))
    lattice_file_name = sat_temp_dir.joinpath("lat_second_pass.ark")
    lattice_output_file_name = sat_temp_dir.joinpath("lat_second_pass_rescore.ark")
    utt2spk = KaldiMapping()
    utt2spk["1-1"] = "1"
    decoder = LmRescorer(gc.g_fst)
    lattice_archive = LatticeArchive(lattice_file_name)
    decoder.export_lattices(lattice_output_file_name, lattice_archive, gc.g_carpa)
    assert lattice_output_file_name.exists()



"""Classes for storing and processing features"""
from __future__ import annotations

import os
import pathlib
import typing

from _kalpy import feat, transform
from _kalpy.matrix import DoubleMatrix, FloatMatrix
from _kalpy.util import (
    RandomAccessBaseDoubleMatrixReader,
    RandomAccessBaseFloatMatrixReader,
    RandomAccessBaseFloatVectorReader,
)
from kalpy.data import KaldiMapping, MatrixArchive
from kalpy.utils import generate_read_specifier, read_kaldi_object


class FeatureArchive:
    def __init__(
        self,
        file_name: typing.Union[pathlib.Path, str],
        utt2spk: KaldiMapping = None,
        cmvn_file_name: typing.Union[pathlib.Path, str] = None,
        lda_mat_file_name: typing.Union[pathlib.Path, str] = None,
        transform_file_name: typing.Union[pathlib.Path, str] = None,
        vad_file_name: typing.Union[pathlib.Path, str] = None,
        use_sliding_cmvn: bool = False,
        cmvn_norm_vars: bool = False,
        cmvn_reverse: bool = False,
        deltas: bool = False,
        splices: bool = False,
        splice_frames: int = 3,
        subsample_n: int = 0,
        sliding_cmvn_window: int = 300,
        sliding_cmvn_center_window: bool = True,
        double: bool = False,
        callback: typing.Callable = None,
    ):
        self.cmvn_reader = None
        self.transform_reader = None
        self.vad_reader = None
        if not os.path.exists(file_name):
            raise OSError(f"Specified file does not exist: {file_name}")
        self.file_name = str(file_name)
        self.archive = MatrixArchive(file_name, double=double)
        self.utt2spk = utt2spk
        self.double = double
        self.subsample_n = subsample_n
        self.callback = callback

        self.use_sliding_cmvn = use_sliding_cmvn
        self.cmvn_norm_vars = cmvn_norm_vars
        self.cmvn_reverse = cmvn_reverse

        self.sliding_cmvn_options = feat.SlidingWindowCmnOptions()
        self.sliding_cmvn_options.cmn_window = sliding_cmvn_window
        self.sliding_cmvn_options.center = sliding_cmvn_center_window
        self.sliding_cmvn_options.normalize_variance = cmvn_norm_vars

        self.delta_options = feat.DeltaFeaturesOptions()
        self.splice_frames = splice_frames
        self.use_deltas = deltas
        self.use_splices = splices
        self.cmvn_read_specifier = None
        if cmvn_file_name:
            self.cmvn_read_specifier = generate_read_specifier(cmvn_file_name)
            self.cmvn_reader = RandomAccessBaseDoubleMatrixReader(self.cmvn_read_specifier)

        self.lda_mat_file_name = None
        self.lda_mat = None
        if lda_mat_file_name:
            self.use_splices = True
            self.use_deltas = False
            self.lda_mat_file_name = str(lda_mat_file_name)
            self.lda_mat = read_kaldi_object(FloatMatrix, self.lda_mat_file_name)
        self.transform_file_name = transform_file_name
        self.transform_read_specifier = None
        if transform_file_name:
            self.transform_read_specifier = generate_read_specifier(transform_file_name)
            self.transform_reader = RandomAccessBaseFloatMatrixReader(
                self.transform_read_specifier
            )

        self.vad_read_specifier = None
        if vad_file_name:
            self.vad_read_specifier = generate_read_specifier(vad_file_name)
            self.vad_reader = RandomAccessBaseFloatVectorReader(self.vad_read_specifier)
        self.current_speaker = None
        self.trans = None
        self.cmvn_stats = None

    def __del__(self):
        self.close()

    def close(self):
        if getattr(self, "archive", None) is not None and self.archive.random_reader.IsOpen():
            self.archive.random_reader.Close()
        if self.cmvn_reader is not None and self.cmvn_reader.IsOpen():
            self.cmvn_reader.Close()
        if self.transform_reader is not None and self.transform_reader.IsOpen():
            self.transform_reader.Close()
        if self.vad_reader is not None and self.vad_reader.IsOpen():
            self.vad_reader.Close()

    def __iter__(self) -> typing.Generator[typing.Tuple[str, FloatMatrix]]:
        """Iterate over the utterance features in the archive"""
        reader = self.archive.sequential_reader
        self.current_speaker = None
        self.trans = None
        self.cmvn_stats = None
        try:
            while not reader.Done():
                utt = reader.Key()
                feats = reader.Value()
                if self.utt2spk is not None:
                    speaker = self.utt2spk[utt]
                else:
                    speaker = None
                if self.current_speaker != speaker:
                    if self.cmvn_reader and speaker is not None:
                        if not self.cmvn_reader.HasKey(speaker):
                            raise Exception(
                                f"Could not find key {speaker} in {self.cmvn_read_specifier}"
                            )
                        self.cmvn_stats = self.cmvn_reader.Value(speaker)
                    if self.transform_reader is not None:
                        if self.transform_reader.HasKey(speaker):
                            self.trans = self.transform_reader.Value(speaker)
                        else:
                            self.trans = None
                    self.current_speaker = speaker
                # Apply CMVN
                if self.cmvn_stats is not None:
                    feats = transform.apply_cmvn(
                        feats,
                        self.cmvn_stats,
                        reverse=self.cmvn_reverse,
                        norm_vars=self.cmvn_norm_vars,
                    )
                elif self.use_sliding_cmvn:
                    feats = feat.sliding_window_cmn(self.sliding_cmvn_options, feats)

                # Deltas or splices
                if self.use_deltas:
                    feats = feat.compute_deltas(self.delta_options, feats)
                elif self.use_splices:
                    feats = feat.splice_frames(feats, self.splice_frames, self.splice_frames)
                    if self.lda_mat is not None:
                        feats = transform.apply_transform(feats, self.lda_mat)

                # Speaker adapted features
                if self.trans is not None:
                    feats = transform.apply_transform(feats, self.trans)

                # Subsampling
                if self.vad_reader is not None and self.vad_reader.HasKey(utt):
                    vad = self.vad_reader.Value(utt)
                    feats = feat.select_voiced_frames(feats, vad)
                if self.subsample_n and self.subsample_n > 0:
                    feats = feat.subsample_feats(feats, n=self.subsample_n)
                yield utt, feats
                reader.Next()
        finally:
            reader.Close()

    def __getitem__(self, item: str) -> FloatMatrix:
        """Get features for a particular key from the archive file"""
        item = str(item)
        if not self.archive.random_reader.HasKey(item):
            raise KeyError(f"No key {item} found in {self.archive.file_name}")
        feats = self.archive.random_reader.Value(item)
        if self.utt2spk is not None:
            speaker = self.utt2spk[item]
        else:
            speaker = None
        if self.current_speaker != speaker:
            if self.cmvn_reader and speaker is not None:
                if not self.cmvn_reader.HasKey(speaker):
                    raise Exception(f"Could not find key {speaker} in {self.cmvn_read_specifier}")
                self.cmvn_stats = self.cmvn_reader.Value(speaker)
            if self.transform_reader is not None:
                if self.transform_reader.HasKey(speaker):
                    self.trans = self.transform_reader.Value(speaker)
                else:
                    self.trans = None
            self.current_speaker = speaker
        # Apply CMVN
        if self.cmvn_stats is not None:
            feats = transform.apply_cmvn(
                feats, self.cmvn_stats, reverse=self.cmvn_reverse, norm_vars=self.cmvn_norm_vars
            )
        elif self.use_sliding_cmvn:
            feats = feat.sliding_window_cmn(self.sliding_cmvn_options, feats)

        # Deltas or splices
        if self.use_deltas:
            feats = feat.compute_deltas(self.delta_options, feats)
        elif self.use_splices:
            feats = feat.splice_frames(feats, self.splice_frames, self.splice_frames)
            if self.lda_mat_file_name:
                feats = transform.apply_transform(feats, self.lda_mat)

        # Speaker adapted features
        if self.trans is not None:
            feats = transform.apply_transform(feats, self.trans)

        # Subsampling
        if self.vad_reader is not None and self.vad_reader.HasKey(item):
            vad = self.vad_reader.Value(item)
            feats = feat.select_voiced_frames(feats, vad)
        if self.subsample_n and self.subsample_n > 0:
            feats = feat.subsample_feats(feats, n=self.subsample_n)
        return feats



import logging
import os
import sys

import pytest
import pywrapfst

from kalpy.decoder.decode_graph import DecodeGraphCompiler
from kalpy.decoder.training_graphs import TrainingGraphCompiler
from kalpy.fstext.lexicon import LexiconCompiler
from kalpy.fstext.utils import kaldi_to_pynini

logger = logging.getLogger("kalpy.graphs")
handler = logging.StreamHandler(sys.stdout)
logger.addHandler(handler)


@pytest.mark.order(1)
def test_training_graphs(
    mono_tree_path, mono_model_path, dictionary_path, mono_temp_dir, acoustic_corpus_text
):
    lc = LexiconCompiler(position_dependent_phones=False)
    lc.load_pronunciations(dictionary_path)
    lc.fst.write(str(mono_temp_dir.joinpath("lexicon.fst")))
    gc = TrainingGraphCompiler(mono_model_path, mono_tree_path, lc)
    graph = kaldi_to_pynini(gc.compile_fst(acoustic_corpus_text))
    assert graph.num_states() > 0
    assert graph.start() != pywrapfst.NO_STATE_ID
    output_file_name = mono_temp_dir.joinpath("fsts.ark")
    gc.export_graphs(output_file_name, [("1-1", acoustic_corpus_text)])
    assert output_file_name.exists()
    os.remove(output_file_name)
    gc.export_graphs(output_file_name, [("1-1", acoustic_corpus_text)], write_scp=True)
    assert output_file_name.exists()
    assert output_file_name.with_suffix(".scp").exists()


@pytest.mark.order(1)
def test_training_graphs_sat(
    sat_tree_path,
    sat_model_path,
    sat_dictionary_path,
    sat_temp_dir,
    acoustic_corpus_text,
    sat_phones,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones)
    lc.load_pronunciations(sat_dictionary_path)
    lc.fst.write(str(sat_temp_dir.joinpath("L_debug.fst")))
    lc.word_table.write_text(str(sat_temp_dir.joinpath("words.txt")))
    lc.phone_table.write_text(str(sat_temp_dir.joinpath("phones.txt")))
    gc = TrainingGraphCompiler(sat_model_path, sat_tree_path, lc)
    graph = kaldi_to_pynini(gc.compile_fst(acoustic_corpus_text))
    assert graph.num_states() > 0
    assert graph.start() != pywrapfst.NO_STATE_ID
    graph.write(str(sat_temp_dir.joinpath("LG_debug.fst")))

    output_file_name = sat_temp_dir.joinpath("fsts.ark")
    gc.export_graphs(output_file_name, [("1-1", acoustic_corpus_text)])
    assert output_file_name.exists()
    os.remove(output_file_name)
    gc.export_graphs(output_file_name, [("1-1", acoustic_corpus_text)], write_scp=True)
    assert output_file_name.exists()
    assert output_file_name.with_suffix(".scp").exists()


@pytest.mark.order(1)
def test_decoding_model_sat(
    sat_tree_path,
    sat_model_path,
    sat_dictionary_path,
    sat_temp_dir,
    acoustic_corpus_text,
    sat_phones,
    lm_path,
):
    lc = LexiconCompiler(position_dependent_phones=False, phones=sat_phones, disambiguation=True)
    lc.load_pronunciations(sat_dictionary_path)
    lc.fst.write(str(sat_temp_dir.joinpath("L_debug.fst")))
    lc.word_table.write_text(str(sat_temp_dir.joinpath("words.txt")))
    lc.phone_table.write_text(str(sat_temp_dir.joinpath("phones.txt")))
    gc = DecodeGraphCompiler(sat_model_path, sat_tree_path, lc)
    gc.compile_hclg_fst(lm_path)
    hclg = gc.hclg_fst
    assert hclg.Type() == "const"
    assert hclg.NumStates() > 0
    assert hclg.Start() != pywrapfst.NO_STATE_ID

    output_file_name = sat_temp_dir.joinpath("hclg.fst")
    gc.export_hclg(lm_path, output_file_name)
    gc.export_g(sat_temp_dir.joinpath("g.fst"))
    assert output_file_name.exists()
    gc.load_from_file(output_file_name)
    assert gc.hclg_fst.Type() == "const"
    assert gc.hclg_fst.Start() != pywrapfst.NO_STATE_ID
    assert gc.hclg_fst.NumStates() == hclg.NumStates()



"""Classes for utterance data"""
from __future__ import annotations

import typing
from dataclasses import dataclass

from _kalpy.feat import (
    DeltaFeaturesOptions,
    SlidingWindowCmnOptions,
    compute_deltas,
    paste_feats,
    sliding_window_cmn,
    splice_frames,
)
from _kalpy.matrix import DoubleMatrix, FloatMatrix
from _kalpy.transform import ApplyCmvn, apply_transform
from kalpy.data import Segment
from kalpy.feat.cmvn import CmvnComputer

if typing.TYPE_CHECKING:
    from kalpy.feat.mfcc import MfccComputer
    from kalpy.models import AcousticModel


@dataclass
class Utterance:
    segment: Segment
    transcript: str
    cmvn_string: typing.Optional[str] = None
    fmllr_string: typing.Optional[str] = None
    mfccs: typing.Optional[FloatMatrix] = None

    def generate_mfccs(self, mfcc_computer: MfccComputer):
        if self.mfccs is None:
            self.mfccs = mfcc_computer.compute_mfccs_for_export(self.segment, compress=False)

    def apply_cmvn(
        self, cmvn: DoubleMatrix, sliding_cmvn: bool = False, cmvn_norm_vars: bool = False
    ):

        if not sliding_cmvn:
            ApplyCmvn(cmvn, cmvn_norm_vars, self.mfccs)
        else:
            sliding_cmvn_options = SlidingWindowCmnOptions()
            sliding_cmvn_options.cmn_window = 300
            sliding_cmvn_options.center = True
            sliding_cmvn_options.normalize_variance = cmvn_norm_vars
            self.mfccs = sliding_window_cmn(sliding_cmvn_options, self.mfccs)

    def generate_features(
        self,
        acoustic_model: AcousticModel,
        cmvn: typing.Optional[DoubleMatrix] = None,
        fmllr_trans: FloatMatrix = None,
        uses_speaker_adaptation: bool = True,
        uses_splices: bool = False,
        uses_deltas: bool = True,
        splice_context: int = 3,
    ):
        if acoustic_model.lda_mat is not None:
            uses_splices = True
            uses_deltas = False
        if self.mfccs is None:
            self.generate_mfccs(acoustic_model.mfcc_computer)
            if cmvn is None:
                cmvn_computer = CmvnComputer()
                cmvn = cmvn_computer.compute_cmvn_from_features([self.mfccs])
            self.apply_cmvn(cmvn)
        feats = self.mfccs

        if acoustic_model.pitch_computer is not None:
            pitch = acoustic_model.pitch_computer.compute_pitch_for_export(
                self.segment, compress=False
            )
            feats = paste_feats([feats, pitch], 1)
        if uses_splices:
            feats = splice_frames(feats, splice_context, splice_context)
            if acoustic_model.lda_mat is not None:
                feats = apply_transform(feats, acoustic_model.lda_mat)
        elif uses_deltas:
            delta_options = DeltaFeaturesOptions()
            feats = compute_deltas(delta_options, feats)
        if uses_speaker_adaptation and fmllr_trans is not None:
            feats = apply_transform(feats, fmllr_trans)
        return feats



def test_import():
    import _kalpy

    print(_kalpy.__version__)
    import _kalpy.chain
    import _kalpy.cudamatrix
    import _kalpy.decoder
    import _kalpy.feat
    import _kalpy.fstext
    import _kalpy.gmm
    import _kalpy.hmm
    import _kalpy.ivector
    import _kalpy.kws
    import _kalpy.lat
    import _kalpy.lm
    import _kalpy.matrix
    import _kalpy.tree
    import _kalpy.util
